# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.common.reference
   :synopsis: URN related objects

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>

>>> from MyCapytain.common.reference import (URN, Reference, Citation)

"""
from __future__ import unicode_literals

from collections import defaultdict
from past.builtins import basestring
from builtins import range, object
import re


REFSDECL_SPLITTER = re.compile("/+[a-zA-Z0-9:\[\]@=\\\{\$'\"\.\s]+")
REFSDECL_REPLACER = re.compile("\$[0-9]+")
SUBREFERENCE = re.compile("(\w*)\[{0,1}([0-9]*)\]{0,1}", re.UNICODE)
REFERENCE_REPLACER = re.compile("(@[a-zA-Z0-9:]+){1}(=){1}([\\\$'\"?0-9]{3,6})")


class Reference(object):

    """ A reference object giving informations

    :param reference: Passage Reference part of a Urn
    :type reference: basestring

    :Example:
        >>>    a = Reference(reference="1.1@Achiles[1]-1.2@Zeus[1]")
        >>>    b = Reference(reference="1.1")

    .. automethod:: __str__
    .. automethod:: __eq__
    .. automethod:: __getitem__
    .. automethod:: __setitem__
    """

    def __init__(self, reference):
        self.reference = reference
        self.parsed = self.__parse(reference)

    def __eq__(self, other):
        """ Equality checker for Reference object
        
        :param other: An object to be checked against
        :rtype: boolean
        :returns: Equality between other and self

        :Example:
            >>>    a = Reference(reference="1.1@Achiles[1]-1.2@Zeus[1]")
            >>>    b = Reference(reference="1.1")
            >>>    c = Reference(reference="1.1")
            >>>    (a == b) == False
            >>>    (c == b) == True
        """
        return (isinstance(other, self.__class__)
                and self.reference == str(other))

    @property
    def parent(self):
        """

        :return:
        """
        if len(self.parsed[0][1]) == 1 and len(self.parsed[1][1]) <= 1:
            return None
        else:
            if len(self.parsed[0][1]) > 1 and len(self.parsed[1][1]) == 0:
                return "{0}{1}".format(
                    ".".join(list(self.parsed[0][1])[0:-1]),
                    self.parsed[0][3] or ""
                )
            elif len(self.parsed[0][1]) > 1 and len(self.parsed[1][1]) > 1:
                first = list(self.parsed[0][1])[0:-1]
                last = list(self.parsed[1][1])[0:-1]
                if first == last and self.parsed[1][3] is None \
                    and self.parsed[0][3] is None:
                    return ".".join(first)
                else:
                    return "{0}{1}-{2}{3}".format(
                        ".".join(first),
                        self.parsed[0][3] or "",
                        ".".join(list(self.parsed[1][1])[0:-1]),
                        self.parsed[1][3] or ""
                    )

    def __str__(self):
        """ Return full reference in string format
        
        :rtype: basestring
        :returns: String representation of Reference Object

        :Example:
            >>>    a = Reference(reference="1.1@Achiles[1]-1.2@Zeus[1]")
            >>>    b = Reference(reference="1.1")
            >>>    str(a) == "1.1@Achiles[1]-1.2@Zeus[1]"
            >>>    str(b) == "1.1"
        """
        return self.reference

    def __getitem__(self, key):
        """ Return part of or full passage reference 
        
        Available keys :
            - *1 | start* : First part of the reference
            - *2 | start_list* : Reference start parsed into a list
            - *3 | start_sub* : Subreference start parsed into a tuple
            - *4 | end* : Last part of the reference
            - *5 | end_list* : Reference start parsed into a list
            - *6 | end_sub* : Subreference end parsed into a tuple
            - *default* : full string reference 

        :param key: Identifier of the part to return
        :type key: basestring or int
        :rtype: basestring or List.<int> or None or Tuple.<string>
        :returns: Desired part of the passage reference

        :Example:
            >>>    a = Reference(reference="1.1@Achiles[1]-1.2@Zeus[1]")
            >>>    print(a[1]) # "1.1@Achiles[1]"
            >>>    print(a["start_list"]) # ("1", "1")
            >>>    print(a[6]) # ("Zeus", "1")
            >>>    print(a[7]) # "1.1@Achiles[1]-1.2@Zeus[1]"
        """
        if key == 1 or key == "start":
            return self.parsed[0][0]
        elif key == 4 or key == "end":
            return self.parsed[1][0]
        elif key == 2 or key == "start_list":
            return self.parsed[0][1]
        elif key == 5 or key == "end_list":
            return self.parsed[1][1]
        elif key == 3 or key == "start_sub":
            return self.parsed[0][2]
        elif key == 6 or key == "end_sub":
            return self.parsed[1][2]
        else:
            return self.reference

    def __model(self):
        """ 3-Tuple model for references
        
            First element is full text reference,
            Second is list of passage identifiers
            Third is subreference

        :rtype: Tuple
        :returns: An empty tuple to model data
        """
        return [None, [], None, None]

    def __regexp(self, subreference):
        """ Split components of subreference 
        
        :param subreference: A subreference
        :type subreference: basestring
        :rtype: List.<Tuple>
        :returns: List where first element is a tuple representing different components
        """
        return SUBREFERENCE.findall(subreference)[0]

    def __parse(self, reference):
        """ Parse references informations
        
        """

        ref = reference.split("-")
        element = [self.__model(), self.__model()]
        for i in range(0, len(ref)):
            r = ref[i]
            element[i][0] = r
            subreference = r.split("@")
            if len(subreference) == 2:
                element[i][2] = self.__regexp(subreference[1])
                element[i][3] = "@" + subreference[1]
                r = subreference[0]
            element[i][1] = r.split(".")
            element[i] = tuple(element[i])
        return tuple(element)


class URN(object):

    """ A URN object giving all useful sections 

    :param urn: A CTS URN
    :type urn: basestring

    :Example:
        >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1")

    .. automethod:: __len__
    .. automethod:: __gt__
    .. automethod:: __lt__
    .. automethod:: __eq__
    .. automethod:: __str__
    .. automethod:: __getitem__
    """
        

    __order = [
        "full",
        "urn_namespace",
        "cts_namespace",
        "textgroup",
        "work",
        "text",
        "passage",
        "reference"  # Reference is a more complex object
    ]

    def __init__(self, urn):
        self.urn = urn
        self.parsed = self.__parse(self.urn)

    def __len__(self):
        """ Returns the len of the URN

        :rtype: int
        :returns: Length of the URN

        .. warning:: Does not take into account the passage !

        :Example:
            >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1") 
            >>>    print(len(a)) # 


        """
        
        items = [key for key in self.parsed if key not in ["passage", "reference", "full"] ]
        return len(items)

    def __gt__(self, other):
        """ Allows for greater comparison

        :param other: Comparison object
        :type other: URN
        :rtype: boolean
        :returns: Indicator of bigger size

        .. warning:: Does not take into account the passage !

        :Example:
            >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1") 
            >>>    b = URN(urn="urn:cts:latinLit:phi1294.phi002:1.1") 
            >>>    (a > b) == True # 
        """
        return len(self) > len(other)

    def __lt__(self, other):
        """ Allows for lower comparison

        :param other: Comparison object
        :type other: URN
        :rtype: boolean
        :returns: Indicator of lower size

        .. warning:: Does not take into account the passage !

        :Example:
            >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1") 
            >>>    b = URN(urn="urn:cts:latinLit:phi1294.phi002:1.1") 
            >>>    (b < a) == True # 
        """
        return len(self) < len(other)

    def __eq__(self, other):
        """ Equality checker for URN object
        
        :param other: An object to be checked against
        :type other: URN
        :rtype: boolean
        :returns: Equality between other and self

        :Example:
            >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1") 
            >>>    b = URN(urn="urn:cts:latinLit:phi1294.phi002:1.1") 
            >>>    (b == a) == False # 
        """
        return (isinstance(other, self.__class__)
                and self.parsed["full"] == other["full"])

    def __str__(self):
        """ Return full initial urn
        
        :rtype: basestring
        :returns: String representation of URN Object

        :Example:
            >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1") 
            >>>    str(a) == "urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1"
        """
        return self.urn

    def __getitem__(self, key):
        """ Returns the urn (int) level or up to (str) level. 
        

        Available keys :
            - *0* :  URN
            - *full* : URN
            - *1* :  Namespace of the urn (cts)
            - *urn_namespace* : URN until the Namespace of the urn 
            - *2* :  CTS Namespace of the URN (e.g. latinLit)
            - *cts_namespace* : URN until the CTS Namespace
            - *3* :  Textgroup of the URN
            - *textgroup* : URN until the Textgroup
            - *4* :  Work of the URN
            - *work* : URN until the Work
            - *5* :  Text of the URN
            - *text* : URN until the Text
            - *6* or *passage*:  Passage of URN

        :param key: Identifier of the wished resource
        :type key: int or basestring
        :rtype: basestring or Reference
        :returns: Part or complete URN
        :warning: *urn:* is not counted as an element !

        :Example:
            >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1") 
            >>>    a["textgroup"] == "urn:cts:latinLit:phi1294"
            >>>    a[3] == "phi1294
        """
        if isinstance(key, int) and key < len(URN.__order):
            return self.parsed[URN.__order[key]]
        elif key == "urn_namespace":
            return ":".join(["urn", self.parsed["urn_namespace"]])
        elif key == "cts_namespace":
            return ":".join([
                "urn",
                self.parsed["urn_namespace"],
                self.parsed["cts_namespace"]])
        elif key == "textgroup" and "textgroup" in self.parsed:
            return ":".join([
                "urn",
                self.parsed["urn_namespace"],
                self.parsed["cts_namespace"],
                self.parsed["textgroup"]
            ])
        elif key == "work" and "work" in self.parsed:
            return ":".join([
                "urn",
                self.parsed["urn_namespace"],
                self.parsed["cts_namespace"],
                ".".join(
                    [
                        self.parsed["textgroup"],
                        self.parsed["work"]
                    ])
            ])
        elif key == "text" and "text" in self.parsed:
            return ":".join([
                "urn",
                self.parsed["urn_namespace"],
                self.parsed["cts_namespace"],
                ".".join(
                    [
                        self.parsed["textgroup"],
                        self.parsed["work"],
                        self.parsed["text"]
                    ])
            ])
        elif key == "passage" and "passage" in self.parsed:
            if "text" in self.parsed:
                return ":".join([
                    "urn",
                    self.parsed["urn_namespace"],
                    self.parsed["cts_namespace"],
                    ".".join(
                        [
                            self.parsed["textgroup"],
                            self.parsed["work"],
                            self.parsed["text"]
                        ]),
                    self.parsed["passage"]
                ])
            else:
                return ":".join([
                    "urn",
                    self.parsed["urn_namespace"],
                    self.parsed["cts_namespace"],
                    ".".join(
                        [
                            self.parsed["textgroup"],
                            self.parsed["work"]
                        ]),
                    self.parsed["passage"]
                ])
        elif key == "start" and "passage" in self.parsed:
            if "text" in self.parsed:
                return ":".join([
                    "urn",
                    self.parsed["urn_namespace"],
                    self.parsed["cts_namespace"],
                    ".".join(
                        [
                            self.parsed["textgroup"],
                            self.parsed["work"],
                            self.parsed["text"]
                        ]),
                    self.parsed["reference"]["start"]
                ])
            else:
                return ":".join([
                    "urn",
                    self.parsed["urn_namespace"],
                    self.parsed["cts_namespace"],
                    ".".join(
                        [
                            self.parsed["textgroup"],
                            self.parsed["work"]
                        ]),
                    self.parsed["reference"]["start"]
                ])
        elif key == "end" and "passage" in self.parsed and self.parsed["reference"]["end"] is not None:
            if "text" in self.parsed:
                return ":".join([
                    "urn",
                    self.parsed["urn_namespace"],
                    self.parsed["cts_namespace"],
                    ".".join([
                        self.parsed["textgroup"],
                        self.parsed["work"],
                        self.parsed["text"]
                    ]),
                    self.parsed["reference"]["end"]
                ])
            else:
                return ":".join([
                    "urn",
                    self.parsed["urn_namespace"],
                    self.parsed["cts_namespace"],
                    ".".join([
                        self.parsed["textgroup"],
                        self.parsed["work"]
                    ]),
                    self.parsed["reference"]["end"]
                ])
        elif key == "full":
            return self.parsed["full"]
        elif key == "reference" and "reference" in self.parsed:
            return self.parsed["reference"]
        else:
            return None

    def __parse(self, urn):
        """ Parse a URN
        

        :param urn: A URN:CTS
        :type urn: basestring
        :rtype: defaultdict.basestring
        :returns: Dictionary representation
        """
        parsed = defaultdict(str)
        parsed["full"] = urn.split("#")[0]
        urn = parsed["full"].split(":")
        if isinstance(urn, list) and len(urn) > 2:
            parsed["urn_namespace"] = urn[1]
            parsed["cts_namespace"] = urn[2]

            if len(urn) == 5:
                parsed["passage"] = urn[4]
                parsed["reference"] = Reference(urn[4])

            if len(urn) >= 4:
                urn = urn[3].split(".")
                if len(urn) >= 1:
                    parsed["textgroup"] = urn[0]
                if len(urn) >= 2:
                    parsed["work"] = urn[1]
                if len(urn) >= 3:
                    parsed["text"] = urn[2]
        else:
            raise ValueError("URN is empty")
        return parsed


class Citation(object):
    """ A citation object gives informations about the scheme 
        
    :param name: Name of the citation (e.g. "book")
    :type name: basestring
    :param xpath: Xpath of the citation (As described by CTS norm)
    :type xpath: basestring
    :param scope: Scope of the citation (As described by CTS norm)
    :type xpath: basestring
    :param refsDecl: refsDecl version
    :type refsDecl: basestring
    :param child: A citation
    :type child: Citation

    .. automethod:: __iter__
    .. automethod:: __len__
    """

    def __init__(self, name=None, xpath=None, scope=None, refsDecl=None, child=None):
        """ Initialize a Citation object
        """
        self.__name = None
        self.__xpath = None
        self.__scope = None
        self.__refsDecl = None
        self.__child = None

        self.name = name
        self.scope = scope
        self.xpath = xpath
        self.refsDecl = refsDecl

        if child is not None:
            self.child = child

    @property
    def name(self): 
        """ Type of the citation represented
        
        :type: basestring
        :Example: Book, Chapter, Textpart, Section, Poem...
        """
        return self.__name

    @name.setter
    def name(self, val):
        self.__name = val

    @property
    def xpath(self):
        """ TextInventory xpath property of a citation (ie. identifier of the last element of the citation)
        
        :type: basestring
        :Example: //tei:l[@n="?"]
        """
        return self.__xpath

    @xpath.setter
    def xpath(self, val):
        if val is not None:
            self.__xpath = val
            self.__upRefsDecl()

    @property
    def scope(self):
        """ TextInventory scope property of a citation (ie. identifier of all element but the last of the citation)
        
        :type: basestring
        :Example: /tei:TEI/tei:text/tei:body/tei:div
        """
        return self.__scope
        
    @scope.setter
    def scope(self, val):
        if val is not None:
            self.__scope = val
            self.__upRefsDecl()

    @property
    def refsDecl(self):
        """ ResfDecl expression of the citation scheme 

        :type: basestring
        :Example: /tei:TEI/tei:text/tei:body/tei:div//tei:l[@n='$1']
        """
        return self.__refsDecl
        
    @refsDecl.setter
    def refsDecl(self, val):
        if val is not None:
            self.__refsDecl = val
            self.__upXpathScope()

    @property
    def child(self):
        """ Child of a citation

        :type: Citation or None
        :Example: Citation.name==poem would have a child Citation.name==line
        """
        return self.__child
        
    @child.setter
    def child(self, val):
        if isinstance(val, self.__class__):
            self.__child = val

    def __upXpathScope(self):
        """ Update xpath and scope property when refsDecl is updated
        
        """
        rd = self.__refsDecl
        matches = REFSDECL_SPLITTER.findall(rd)
        self.__scope = REFSDECL_REPLACER.sub("?", "".join(matches[0:-1]))
        self.__xpath = REFSDECL_REPLACER.sub("?", matches[-1])

    def __upRefsDecl(self):
        """ Update xpath and scope property when refsDecl is updated
        
        """
        if self.__scope is not None and self.__xpath is not None:
            xpath = self.__scope + self.__xpath
            i = xpath.find("?")
            ii = 1
            while i >= 0:
                xpath = xpath[:i] + "$" + str(ii) + xpath[i+1:]
                i = xpath.find("?")
                ii += 1
            self.__refsDecl = xpath

    def __iter__(self):
        """ Iteration method
        
        Loop over the citation childs

        :Example:
            >>>    c = Citation(name="line")
            >>>    b = Citation(name="poem", child=c)
            >>>    a = Citation(name="book", child=b)
            >>>    [e for e in a] == [a, b, c]
            
        """
        e = self
        while e is not None:
            yield e
            if hasattr(e, "child") and e.child is not None:
                e = e.child
            else:
                break

    def __len__(self):
       """ Length method

       :rtype: int
       :returns: Number of nested citations
       """
       return len([item for item in self])

    def fill(self, passage=None, xpath=None):
        """ Fill the xpath with given informations

        :param passage: Passage reference
        :type passage: Reference or lsit
        :param xpath: If set to True, will return the replaced self.xpath value and not the whole self.refsDecl
        :type xpath: Boolean
        :rtype: basestring
        :returns: Xpath to find the passage
        """
        if xpath is True: # Then passage is a string or None
            xpath = self.xpath

            if passage is None:
                replacement = r"\1"
            elif isinstance(passage, basestring):
                replacement = r"\1\2'" + passage + "'"

            return REFERENCE_REPLACER.sub(replacement, xpath)
        else:        
            if isinstance(passage, Reference):
                passage = passage[2]
            passage = iter(passage)    
            return REFERENCE_REPLACER.sub(
                lambda m: REF_REPLACER(m, passage),
                self.refsDecl
            )

def REF_REPLACER(match, passage):
    """ Helper to replace xpath/scope/refsDecl on iteration with passage value

    :param match: A RegExp match
    :type match: re.SRE_MATCH
    :param passage: A list with subreference informations
    :type passage: iter

    :rtype: basestring
    :return: Replaced string
    """
    groups = match.groups()
    ref = next(passage)
    if ref is None:
        return groups[0]
    else:
        return "{1}='{0}'".format(ref, groups[0])


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unittest
import six
from collections import defaultdict
from MyCapytain.common.metadata import Metadata, Metadatum

class TestMetadatum(unittest.TestCase):
    def test_init(self):
        a = Metadatum("title")
        self.assertEqual(a.name, "title")

    def test_set_item(self):
        a = Metadatum("title")
        a["eng"] = "Epigrams"
        self.assertEqual(a["eng"], "Epigrams")
        # It should set a default too
        self.assertEqual(a["fre"], "Epigrams")
        # But not twice
        a["fre"] = "Epigrammes"
        self.assertEqual(a["spa"], "Epigrams")
        self.assertEqual(a["fre"], "Epigrammes")
        # Set tuples with one value
        a[("lat", "grc")] = "Epigrammata"
        self.assertEqual(a["lat"], "Epigrammata")
        self.assertEqual(a["grc"], "Epigrammata")
        # Set tuples with tuples
        a = Metadatum("title")
        a[("lat", "grc")] = ("Epigrammata", "ἐπιγραμματον")
        self.assertEqual(a["lat"], "Epigrammata")
        self.assertEqual(a["grc"], "ἐπιγραμματον")
        # Set error because key is wrong
        with six.assertRaisesRegex(self, TypeError, "Only basestring or tuple instances are accepted as key"):
            a[3.5] = "test"
        with six.assertRaisesRegex(self, ValueError, "Less values than keys detected"):
            a[("lat", "grc")] = ["Epigrammata"]
        

    def test_init_with_children(self):
        a = Metadatum("title", [
                ("eng", "Epigrams"),
                ("fre", "Epigrammes")
            ])
        self.assertEqual(a["fre"], "Epigrammes")
        self.assertEqual(a["eng"], "Epigrams")

    def test_override_default(self):
        a = Metadatum("title", [
                ("eng", "Epigrams"),
                ("fre", "Epigrammes")
            ])
        self.assertEqual(a["spa"], "Epigrams")
        a.setDefault("fre")
        self.assertEqual(a["spa"], "Epigrammes")
        # If key is not available, should raise an error
        with six.assertRaisesRegex(self, ValueError, "Can not set a default to an unknown key"):
            a.setDefault(None)

    def test_len(self):
        a = Metadatum("title", [
                ("eng", "Epigrams"),
                ("fre", "Epigrammes")
            ])
        self.assertEqual(len(a), 2)

    def test_get_item(self):
        a = Metadatum("title", [
                ("eng", "Epigrams"),
                ("fre", "Epigrammes")
            ])
        self.assertEqual(a[0], "Epigrams")
        self.assertEqual(a["eng"], "Epigrams")
        self.assertEqual(a["spa"], "Epigrams")
        self.assertEqual(a[1], "Epigrammes")
        # Access through tuple !
        self.assertEqual(a[(0,1)], ("Epigrams", "Epigrammes"))
        with self.assertRaises(KeyError):
            z = a[2]

        a.default = None
        with self.assertRaises(KeyError):
            z = a["spa"]

    def test_iter(self):
        data = [
                ("eng", "Epigrams"),
                ("fre", "Epigrammes")
            ]
        a = Metadatum("title", [
                ("eng", "Epigrams"),
                ("fre", "Epigrammes")
            ])

        testdata = [(k, v) for k, v in a]
        self.assertEqual(list(a), testdata)

        a["lat"] = "Epigrammata"
        testdata = [(k, v) for k, v in a]
        self.assertEqual(testdata, data + [("lat", "Epigrammata")])

        i = 0
        d=[]
        for k, v in a:
            d.append(k)
            break
        self.assertEqual(d, ["eng"])

        i = 0
        d=[]
        for k, v in a:
            d.append(k)
            if i == 1:
                break
            i += 1
        self.assertEqual(d, ["eng", "fre"])


class TestMetadata(unittest.TestCase):
    def test_init(self):
        a = Metadata()
        self.assertTrue(hasattr(a, "metadata"), True)
        self.assertTrue(isinstance(a.metadata, defaultdict))

        a = Metadata(keys=["title"])
        self.assertTrue(isinstance(a.metadata["title"], Metadatum))

    def test_set(self):
        a = Metadata()
        a["title"] = [("eng", "Epigrams")]
        self.assertEqual(a["title"]["eng"], "Epigrams")
        a[("desc", "label")] = ([("eng", "desc")], [("eng", "lbl"), ("fre", "label")])
        self.assertEqual(a["desc"]["eng"], "desc")
        self.assertEqual(a["label"][("eng", "fre")], ("lbl", "label"))

    def test_get(self):
        a = Metadata()
        m1 = Metadatum("desc", [("eng", "desc")])
        m2 = Metadatum("label", [("eng", "lbl"), ("fre", "label")])
        a[("desc", "label")] = (m1, m2)
        self.assertEqual(a[("desc", "label")], (m1, m2))

        self.assertEqual(a[0], m1)
        with self.assertRaises(KeyError):
            z = a[2]
        with self.assertRaises(KeyError):
            z = a["textgroup"]

        with six.assertRaisesRegex(self, TypeError, "Only basestring or tuple instances are accepted as key"):
            a[3.5] = "test"
        with six.assertRaisesRegex(self, ValueError, "Less values than keys detected"):
            a[("lat", "grc")] = ["Epigrammata"]

    def test_iter(self):
        a = Metadata()
        m1 = Metadatum("desc", [("eng", "desc")])
        m2 = Metadatum("label", [("eng", "lbl"), ("fre", "label")])
        a["desc"] = m1
        a["label"] = m2

        i = 0
        d=[]
        d2=[]
        for k, v in a:
            d.append(k)
            d2.append(v)
            if i == 1:
                break
            i += 1
        self.assertEqual(d, ["desc", "label"])
        self.assertEqual(d2, [m1, m2])

        i = 0
        d=[]
        d2=[]
        for k, v in a:
            d.append(k)
            d2.append(v)
            break
        self.assertEqual(d, ["desc"])
        self.assertEqual(d2, [m1])

        self.assertEqual(list(a), [("desc", m1), ("label", m2)])

    def test_len(self):
        a = Metadata()
        m1 = Metadatum("desc", [("eng", "desc")])
        m2 = Metadatum("label", [("eng", "lbl"), ("fre", "label")])
        a["desc"] = m1
        self.assertEqual(len(a), 1)
        a["label"] = m2
        self.assertEqual(len(a), 2)
        a["z"] = 1.5
        self.assertEqual(len(a), 2)

    def test_add(self):
        """ Test sum of two Metadata objects  """
        a = Metadata()
        m1 = Metadatum("desc", [("eng", "desc")])
        m2 = Metadatum("label", [("eng", "lbl"), ("fre", "label")])
        a["desc"] = m1
        a["label"] = m2

        b = Metadata()
        m3 = Metadatum("desc", [("fre", "Omelette")])
        m4 = Metadatum("title", [("eng", "ttl"), ("fre", "titre")])
        b[("desc", "title")] = (m3, m4)

        c = a + b
        self.assertEqual(len(c), 3)
        self.assertEqual(len(c["desc"]), 2)




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unittest
from MyCapytain.common.reference import URN, Reference, Citation


class TestReferenceImplementation(unittest.TestCase):

    """ Test how reference reacts """

    def test_str_function(self):
        a = Reference("1-1")
        self.assertEqual(str(a), "1-1")

    def test_str_getitem(self):
        a = Reference("1.1@Achilles[0]-1.10@Atreus[3]")
        self.assertEqual(a["any"], "1.1@Achilles[0]-1.10@Atreus[3]")
        self.assertEqual(a["start"], "1.1@Achilles[0]")
        self.assertEqual(a["start_list"], ["1", "1"])
        self.assertEqual(a["start_sub"][0], "Achilles")
        self.assertEqual(a["end"], "1.10@Atreus[3]")
        self.assertEqual(a["end_list"], ["1", "10"])
        self.assertEqual(a["end_sub"][1], "3")
        self.assertEqual(a["end_sub"], ("Atreus", "3"))

    def test_int_getItem(self):
        a = Reference("1.1@Achilles-1.10@Atreus[3]")
        self.assertEqual(a[1], "1.1@Achilles")
        self.assertEqual(a[2], ["1", "1"])
        self.assertEqual(a[3][0], "Achilles")
        self.assertEqual(a[4], "1.10@Atreus[3]")
        self.assertEqual(a[5], ["1", "10"])
        self.assertEqual(a[6][1], "3")
        self.assertEqual(a[6], ("Atreus", "3"))

    def test_Unicode_Support(self):
        a = Reference("1.1@καὶ[0]-1.10@Ἀλκιβιάδου[3]")
        self.assertEqual(a[1], "1.1@καὶ[0]")
        self.assertEqual(a[2], ["1", "1"])
        self.assertEqual(a[3][0], "καὶ")
        self.assertEqual(a[4], "1.10@Ἀλκιβιάδου[3]")
        self.assertEqual(a[5], ["1", "10"])
        self.assertEqual(a[6][1], "3")
        self.assertEqual(a[6], ("Ἀλκιβιάδου", "3"))

    def test_NoWord_Support(self):
        a = Reference("1.1@[0]-1.10@Ἀλκιβιάδου[3]")
        self.assertEqual(a[1], "1.1@[0]")
        self.assertEqual(a[3][0], "")
        self.assertEqual(a[3][1], "0")

    def test_No_End_Support(self):
        a = Reference("1.1@[0]")
        self.assertEqual(a[4], None)
        self.assertEqual(a[1], "1.1@[0]")
        self.assertEqual(a[3][0], "")
        self.assertEqual(a[3][1], "0")

    def test_equality(self):
        a = Reference("1.1@[0]")
        b = Reference("1.1@[0]")
        c = Reference("1.1@[1]")
        d = "1.1@[0]"
        self.assertEqual(a, b)
        self.assertNotEqual(a, c)
        self.assertNotEqual(a, d)

    def test_get_parent(self):
        a = Reference("1.1")
        b = Reference("1")
        c = Reference("1.1-2.3")
        d = Reference("1.1-1.2")
        e = Reference("1.1@Something[0]-1.2@SomethingElse[2]")
        f = Reference("1-2")

        self.assertEqual(str(a.parent), "1")
        self.assertEqual(b.parent, None)
        self.assertEqual(str(c.parent), "1-2")
        self.assertEqual(str(d.parent), "1")
        self.assertEqual(str(e.parent), "1@Something[0]-1@SomethingElse[2]")
        self.assertEqual(f.parent, None)


class TestURNImplementation(unittest.TestCase):

    """ Test how reference reacts """

    def test_str_function(self):
        a = URN("urn:cts:greekLit:tlg0012.tlg001.mth-01:1.1@Achilles-1.10@the[2]")
        self.assertEqual(str(a), "urn:cts:greekLit:tlg0012.tlg001.mth-01:1.1@Achilles-1.10@the[2]")

    def test_int_access(self):
        a = URN("urn:cts:greekLit:tlg0012.tlg001.mth-01:1.1@Achilles-1.10@the[2]")
        self.assertEqual(a[0], "urn:cts:greekLit:tlg0012.tlg001.mth-01:1.1@Achilles-1.10@the[2]")
        self.assertEqual(a[1], "cts")
        self.assertEqual(a[2], "greekLit")
        self.assertEqual(a[3], "tlg0012")
        self.assertEqual(a[4], "tlg001")
        self.assertEqual(a[5], "mth-01")
        self.assertEqual(a[6], "1.1@Achilles-1.10@the[2]")
        self.assertEqual(a[7], Reference("1.1@Achilles-1.10@the[2]"))

    def test_str_access(self):
        a = URN("urn:cts:greekLit:tlg0012.tlg001.mth-01:1.1@Achilles-1.10@the[2]")
        self.assertEqual(a["full"], "urn:cts:greekLit:tlg0012.tlg001.mth-01:1.1@Achilles-1.10@the[2]")
        self.assertEqual(a["urn_namespace"], "urn:cts")
        self.assertEqual(a["cts_namespace"], "urn:cts:greekLit")
        self.assertEqual(a["textgroup"], "urn:cts:greekLit:tlg0012")
        self.assertEqual(a["work"], "urn:cts:greekLit:tlg0012.tlg001")
        self.assertEqual(a["text"], "urn:cts:greekLit:tlg0012.tlg001.mth-01")
        self.assertEqual(a["passage"], "urn:cts:greekLit:tlg0012.tlg001.mth-01:1.1@Achilles-1.10@the[2]")
        self.assertEqual(a["reference"], Reference("1.1@Achilles-1.10@the[2]"))
        self.assertEqual(a["start"], "urn:cts:greekLit:tlg0012.tlg001.mth-01:1.1@Achilles")
        self.assertEqual(a["end"], "urn:cts:greekLit:tlg0012.tlg001.mth-01:1.10@the[2]")

    def test_equality(self):
        a = URN("urn:cts:greekLit:tlg0012.tlg001.mth-01:1.1@Achilles-1.10@the[2]")
        b = URN("urn:cts:greekLit:tlg0012.tlg001.mth-01:1.1@Achilles-1.10@the[2]")
        c = URN("urn:cts:greekLit:tlg0012.tlg001.mth-01:1.1@Achilles-1.10@the[3]")
        d = "urn:cts:greekLit:tlg0012.tlg001.mth-01:1.1@Achilles-1.10@the[2]"
        self.assertEqual(a,b)
        self.assertNotEqual(a,c)
        self.assertNotEqual(a,d)

    def test_full_emptiness(self):
        a = URN("urn:cts:greekLit")
        self.assertEqual(a["full"], "urn:cts:greekLit")
        self.assertEqual(a["urn_namespace"], "urn:cts")
        self.assertEqual(a["cts_namespace"], "urn:cts:greekLit")
        self.assertIsNone(a["textgroup"])
        self.assertIsNone(a["work"])
        self.assertIsNone(a["text"])
        self.assertIsNone(a["passage"])
        self.assertIsNone(a["reference"])
        self.assertIsNone(a["start"])
        self.assertIsNone(a["end"])

    def test_from_textgroup_emptiness(self):
        a = URN("urn:cts:greekLit:textgroup")
        self.assertEqual(a["full"], "urn:cts:greekLit:textgroup")
        self.assertEqual(a["urn_namespace"], "urn:cts")
        self.assertEqual(a["cts_namespace"], "urn:cts:greekLit")
        self.assertEqual(a["textgroup"], "urn:cts:greekLit:textgroup")
        self.assertIsNone(a["work"])
        self.assertIsNone(a["text"])
        self.assertIsNone(a["passage"])
        self.assertIsNone(a["reference"])
        self.assertIsNone(a["start"])
        self.assertIsNone(a["end"])


    def test_from_work_emptiness(self):
        a = URN("urn:cts:greekLit:textgroup.work")
        self.assertEqual(a["full"], "urn:cts:greekLit:textgroup.work")
        self.assertEqual(a["urn_namespace"], "urn:cts")
        self.assertEqual(a["cts_namespace"], "urn:cts:greekLit")
        self.assertEqual(a["textgroup"], "urn:cts:greekLit:textgroup")
        self.assertEqual(a["work"], "urn:cts:greekLit:textgroup.work")
        self.assertIsNone(a["text"])
        self.assertIsNone(a["passage"])
        self.assertIsNone(a["reference"])
        self.assertIsNone(a["start"])
        self.assertIsNone(a["end"])

    def test_from_text_emptiness(self):
        a = URN("urn:cts:greekLit:textgroup.work.text")
        self.assertEqual(a["full"], "urn:cts:greekLit:textgroup.work.text")
        self.assertEqual(a["urn_namespace"], "urn:cts")
        self.assertEqual(a["cts_namespace"], "urn:cts:greekLit")
        self.assertEqual(a["textgroup"], "urn:cts:greekLit:textgroup")
        self.assertEqual(a["work"], "urn:cts:greekLit:textgroup.work")
        self.assertEqual(a["text"], "urn:cts:greekLit:textgroup.work.text")
        self.assertIsNone(a["passage"])
        self.assertIsNone(a["reference"])
        self.assertIsNone(a["start"])
        self.assertIsNone(a["end"])

    def test_no_end_text_emptiness(self):
        a = URN("urn:cts:greekLit:textgroup.work.text:1")
        self.assertEqual(a["full"], "urn:cts:greekLit:textgroup.work.text:1")
        self.assertEqual(a["urn_namespace"], "urn:cts")
        self.assertEqual(a["cts_namespace"], "urn:cts:greekLit")
        self.assertEqual(a["textgroup"], "urn:cts:greekLit:textgroup")
        self.assertEqual(a["work"], "urn:cts:greekLit:textgroup.work")
        self.assertEqual(a["text"], "urn:cts:greekLit:textgroup.work.text")
        self.assertEqual(a["passage"], "urn:cts:greekLit:textgroup.work.text:1")
        self.assertEqual(a["reference"], Reference("1"))
        self.assertEqual(a["start"], "urn:cts:greekLit:textgroup.work.text:1")
        self.assertIsNone(a["end"])

    def test_missing_text_in_passage_emptiness(self):
        a = URN("urn:cts:greekLit:textgroup.work:1-2")
        self.assertEqual(a["full"], "urn:cts:greekLit:textgroup.work:1-2")
        self.assertEqual(a["urn_namespace"], "urn:cts")
        self.assertEqual(a["cts_namespace"], "urn:cts:greekLit")
        self.assertEqual(a["textgroup"], "urn:cts:greekLit:textgroup")
        self.assertEqual(a["work"], "urn:cts:greekLit:textgroup.work")
        self.assertIsNone(a["text"])
        self.assertEqual(a["passage"], "urn:cts:greekLit:textgroup.work:1-2")
        self.assertEqual(a["reference"], Reference("1-2"))
        self.assertEqual(a["start"], "urn:cts:greekLit:textgroup.work:1")
        self.assertEqual(a["end"], "urn:cts:greekLit:textgroup.work:2")

    def test_warning_on_empty(self):
        with self.assertRaises(ValueError):
            a = URN("urn:cts")

    def test_len(self):
        a = URN("urn:cts:greekLit")
        self.assertEqual(len(a), 2)

    def test_greater(self):
        a = URN("urn:cts:greekLit")
        b = URN("urn:cts:greekLit:textgroup")
        self.assertGreater(b, a)

    def test_lower(self):
        a = URN("urn:cts:greekLit")
        b = URN("urn:cts:greekLit:textgroup")
        self.assertEqual(a < b, True)


class TestCitation(unittest.TestCase):
    """ Test the citation object """
    def test_updateRefsdecl(self):
        c = Citation(
            name="line",
            scope="/TEI/text/body/div/div[@n=\"?\"]",
            xpath="//l[@n=\"?\"]"
        )
        self.assertEqual(c.refsDecl, "/TEI/text/body/div/div[@n=\"$1\"]//l[@n=\"$2\"]")

    def test_updateRefsdecl_and(self):
        c = Citation(
            name="line",
            refsDecl="/TEI/text/body/div/div[@n=\"?\" and @subtype='edition']/l[@n=\"?\"]"
        )
        self.assertEqual(c.refsDecl, "/TEI/text/body/div/div[@n=\"?\" and @subtype='edition']/l[@n=\"?\"]")

    def test_updateScopeXpath(self):
        c = Citation(
            name="line",
            refsDecl="/TEI/text/body/div/div[@n=\"$1\"]//l[@n=\"$2\"]"
        )
        self.assertEqual(c.scope, "/TEI/text/body/div/div[@n=\"?\"]")
        self.assertEqual(c.xpath, "//l[@n=\"?\"]")

    def test_name(self):
        c = Citation(
            name="line"
        )
        self.assertEqual(c.name, "line")

    def test_child(self):
        c = Citation(
            name="line"
        )
        b = Citation(
            name="poem",
            child=c
        )
        self.assertEqual(b.child, c)

    def test_iter(self):
        c = Citation(
            name="line"
        )
        b = Citation(
            name="poem",
            child=c
        )
        a = Citation(
            name="book",
            child=b
        )
        self.assertEqual([e for e in a], [a, b, c])

    def test_len(self):
        c = Citation(
            name="line"
        )
        b = Citation(
            name="poem",
            child=c
        )
        a = Citation(
            name="book",
            child=b
        )
        self.assertEqual(len(a), 3)

    def test_fill(self):
        c = Citation(
            name="line",
            scope="/TEI/text/body/div/div[@n=\"?\"]",
            xpath="//l[@n=\"?\"]"
        )
        self.assertEqual(c.fill(Reference("1.2")), "/TEI/text/body/div/div[@n='1']//l[@n='2']")
        self.assertEqual(c.fill(Reference("1.1")), "/TEI/text/body/div/div[@n='1']//l[@n='1']")
        self.assertEqual(c.fill("1", xpath=True), "//l[@n='1']")
        self.assertEqual(c.fill("2", xpath=True), "//l[@n='2']")
        self.assertEqual(c.fill(None, xpath=True), "//l[@n]")
        self.assertEqual(c.fill([None, None]), "/TEI/text/body/div/div[@n]//l[@n]")


import unittest
from mock import patch
from MyCapytain.endpoints.cts5 import *


class TestEndpointsCTS5(unittest.TestCase):

    """ Test CTS5 Endpoint request making """

    def setUp(self):
        self.cts = CTS("http://domainname.com/rest/cts")

    def test_request_CTS_getCapabilities_arguments(self):
        """ Tests that methods getCapabilities maps correctly to request"""
        with patch('requests.get') as patched_get:
            self.cts.getCapabilities(inventory="inventory")
            patched_get.assert_called_once_with(
                "http://domainname.com/rest/cts", params={
                    "inv": "inventory",
                    "request": "GetCapabilities"
                }
            )

    def test_request_CTS_getValidReff_arguments(self):
        """ Tests that methods getValidReff maps correctly to request"""
        with patch('requests.get') as patched_get:
            self.cts.getValidReff(urn="urn", inventory="inventory", level=1)
            patched_get.assert_called_once_with(
                "http://domainname.com/rest/cts", params={
                    "inv": "inventory",
                    "request": "GetValidReff",
                    "level": "1",
                    "urn": "urn"
                }
            )
        with patch('requests.get') as patched_get:
            self.cts.getValidReff(urn="urn", inventory="inventory")
            patched_get.assert_called_once_with(
                "http://domainname.com/rest/cts", params={
                    "inv": "inventory",
                    "request": "GetValidReff",
                    "urn": "urn"
                }
            )

    def test_request_CTS_getPassage_arguments(self):
        """ Tests that methods getPassage maps correctly to request"""
        with patch('requests.get') as patched_get:
            self.cts.getPassage(urn="urn", inventory="inventory", context=1)
            patched_get.assert_called_once_with(
                "http://domainname.com/rest/cts", params={
                    "inv": "inventory",
                    "request": "GetPassage",
                    "context": "1",
                    "urn": "urn"
                }
            )
        with patch('requests.get') as patched_get:
            self.cts.getPassage(urn="urn", inventory="inventory")
            patched_get.assert_called_once_with(
                "http://domainname.com/rest/cts", params={
                    "inv": "inventory",
                    "request": "GetPassage",
                    "urn": "urn"
                }
            )

    def test_call_with_default(self):
        inv = CTS("http://domainname.com/rest/cts", inventory="annotsrc")
        with patch('requests.get') as patched_get:
            inv.getPassage(urn="urn")
            patched_get.assert_called_once_with(
                "http://domainname.com/rest/cts", params={
                    "inv": "annotsrc",
                    "request": "GetPassage",
                    "urn": "urn"
                }
            )

    def test_request_CTS_getPassagePlus_arguments(self):
        """ Tests that methods getPassagePlus maps correctly to request"""
        with patch('requests.get') as patched_get:
            self.cts.getPassagePlus(
                urn="urn", inventory="inventory", context=1)
            patched_get.assert_called_once_with(
                "http://domainname.com/rest/cts", params={
                    "inv": "inventory",
                    "request": "GetPassagePlus",
                    "context": "1",
                    "urn": "urn"
                }
            )
        with patch('requests.get') as patched_get:
            self.cts.getPassagePlus(urn="urn", inventory="inventory")
            patched_get.assert_called_once_with(
                "http://domainname.com/rest/cts", params={
                    "inv": "inventory",
                    "request": "GetPassagePlus",
                    "urn": "urn"
                }
            )

    def test_request_CTS_getFirstUrn_arguments(self):
        """ Tests that methods getFirstUrn maps correctly to request"""
        with patch('requests.get') as patched_get:
            self.cts.getFirstUrn(urn="urn", inventory="inventory")
            patched_get.assert_called_once_with(
                "http://domainname.com/rest/cts", params={
                    "inv": "inventory",
                    "request": "GetFirstUrn",
                    "urn": "urn"
                }
            )

    def test_request_CTS_getPrevNextUrn_arguments(self):
        """ Tests that methods getPrevNextUrn maps correctly to request"""
        with patch('requests.get') as patched_get:
            self.cts.getPrevNextUrn(urn="urn", inventory="inventory")
            patched_get.assert_called_once_with(
                "http://domainname.com/rest/cts", params={
                    "inv": "inventory",
                    "request": "GetPrevNextUrn",
                    "urn": "urn"
                }
            )

    def test_request_CTS_getLabel_arguments(self):
        """ Tests that methods getLabel maps correctly to request"""
        with patch('requests.get') as patched_get:
            self.cts.getLabel(urn="urn", inventory="inventory")
            patched_get.assert_called_once_with(
                "http://domainname.com/rest/cts", params={
                    "inv": "inventory",
                    "request": "GetLabel",
                    "urn": "urn"
                }
            )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unittest
from six import text_type as str

from MyCapytain.common.utils import xmlparser
from MyCapytain.resources.texts.tei import *
from MyCapytain.common.reference import Reference


class TestTEICitation(unittest.TestCase):
    def test_str(self):
        a = Citation(name="book", xpath="/tei:div[@n=\"?\"]", scope="/tei:TEI/tei:body/tei:text/tei:div")
        self.assertEqual(
            str(a),"<tei:cRefPattern n=\"book\" matchPattern=\"(\\w+)\" replacementPattern=\"#xpath(/tei:TEI/tei:body/tei:text/tei:div/tei:div[@n=\"$1\"])\"><tei:p>This pointer pattern extracts book</tei:p></tei:cRefPattern>"
        )
        b = Citation(name="chapter", xpath="/tei:div[@n=\"?\"]", scope="/tei:TEI/tei:body/tei:text/tei:div/tei:div[@n=\"?\"]")
        self.assertEqual(
            str(b),"<tei:cRefPattern n=\"chapter\" matchPattern=\"(\\w+)\.(\\w+)\" replacementPattern=\"#xpath(/tei:TEI/tei:body/tei:text/tei:div/tei:div[@n=\"$1\"]/tei:div[@n=\"$2\"])\"><tei:p>This pointer pattern extracts chapter</tei:p></tei:cRefPattern>"
        )
        a = Citation()
        self.assertEqual(str(a), "")

    def test_ingest_none(self):
        """ When list of node is empty or when not a list nor a node """
        a = Citation.ingest([])
        self.assertEqual(a, None)
        a = Citation.ingest({})
        self.assertEqual(a, None)

    def test_ingest_multiple(self):
        b = xmlparser("""
<tei:tei xmlns:tei="http://www.tei-c.org/ns/1.0">
<tei:cRefPattern n="line"
             matchPattern="(\\w+).(\\w+).(\\w+)"
             replacementPattern="#xpath(/tei:TEI/tei:text/tei:body/tei:div/tei:div[@n='$1' and @type='section']/tei:div[@n='$2']/tei:l[@n='$3'])">
    <tei:p>This pointer pattern extracts line</tei:p>
</tei:cRefPattern>
<tei:cRefPattern n="poem"
             matchPattern="(\\w+).(\\w+)"
             replacementPattern="#xpath(/tei:TEI/tei:text/tei:body/tei:div/tei:div[@n='$1']/tei:div[@n='$2'])">
    <tei:p>This pointer pattern extracts poem</tei:p>
</tei:cRefPattern>
<tei:cRefPattern n="book"
             matchPattern="(\\w+)"
             replacementPattern="#xpath(/tei:TEI/tei:text/tei:body/tei:div/tei:div[@n='$1'])">
    <tei:p>This pointer pattern extracts book</tei:p>
</tei:cRefPattern>
</tei:tei>
""".replace("\n", "").replace("\s+", " "))

        a = Citation.ingest(b)

        self.assertEqual(
            str(a),
            """<tei:cRefPattern n="book" matchPattern="(\\w+)" replacementPattern="#xpath(/tei:TEI/tei:text/tei:body/tei:div/tei:div[@n=\'$1\'])"><tei:p>This pointer pattern extracts book</tei:p></tei:cRefPattern>"""
        )
        self.assertEqual(
            str(a.child),
            """<tei:cRefPattern n="poem" matchPattern="(\\w+)\.(\\w+)" replacementPattern="#xpath(/tei:TEI/tei:text/tei:body/tei:div/tei:div[@n=\'$1\']/tei:div[@n=\'$2\'])"><tei:p>This pointer pattern extracts poem</tei:p></tei:cRefPattern>"""
        )
        self.assertEqual(
            str(a.child.child),
            """<tei:cRefPattern n="line" matchPattern="(\\w+)\.(\\w+)\.(\\w+)" replacementPattern="#xpath(/tei:TEI/tei:text/tei:body/tei:div/tei:div[@n=\'$1\' and @type=\'section\']/tei:div[@n=\'$2\']/tei:l[@n=\'$3\'])"><tei:p>This pointer pattern extracts line</tei:p></tei:cRefPattern>"""
        )
        self.assertEqual(
            a.child.child.fill(Reference("1.2.3")),
            "/tei:TEI/tei:text/tei:body/tei:div/tei:div[@n=\'1\' and @type=\'section\']/tei:div[@n=\'2\']/tei:l[@n=\'3\']"
        )

    def test_ingest_single(self):
        b = xmlparser("""
<tei:tei xmlns:tei="http://www.tei-c.org/ns/1.0">
<tei:cRefPattern n="line"
             matchPattern="(\\w+).(\\w+).(\\w+)"
             replacementPattern="#xpath(/tei:TEI/tei:text/tei:body/tei:div/tei:div[@n='$1']/tei:div[@n='$2']/tei:l[@n='$3'])">
    <tei:p>This pointer pattern extracts book and poem and line</tei:p>
</tei:cRefPattern>
</tei:tei>
""".replace("\n", "").replace("\s+", " "))
        a = Citation.ingest(b)

        self.assertEqual(
            str(a),
            """<tei:cRefPattern n="line" matchPattern="(\\w+)\.(\\w+)\.(\\w+)" replacementPattern="#xpath(/tei:TEI/tei:text/tei:body/tei:div/tei:div[@n=\'$1\']/tei:div[@n=\'$2\']/tei:l[@n=\'$3\'])"><tei:p>This pointer pattern extracts line</tei:p></tei:cRefPattern>"""
        )

    def test_ingest_single_and(self):
        text = xmlparser("""
<tei:tei xmlns:tei="http://www.tei-c.org/ns/1.0">
    <tei:cRefPattern n="section" matchPattern="(.+)" replacementPattern="#xpath(/tei:TEI/tei:text/tei:body/tei:div[@type='edition']/tei:div[@n='$1' and @type='section'])" />
</tei:tei>
""".replace("\n", "").replace("\s+", " "))
        citation = Citation.ingest(text)
        self.maxDiff = None
        self.assertEqual(
            str(citation),
            """<tei:cRefPattern n="section" matchPattern="(\\w+)" replacementPattern="#xpath(/tei:TEI/tei:text/tei:body/tei:div[@type='edition']/tei:div[@n=\'$1\' and @type='section'])"><tei:p>This pointer pattern extracts section</tei:p></tei:cRefPattern>"""
        )
        self.assertEqual(citation.scope, "/tei:TEI/tei:text/tei:body/tei:div[@type='edition']")
        self.assertEqual(citation.xpath, "/tei:div[@n='?' and @type='section']")
        self.assertEqual(citation.fill("1"), "/tei:TEI/tei:text/tei:body/tei:div[@type='edition']/tei:div[@n=\'1\' and @type='section']")


class TestTEIPassage(unittest.TestCase):
    def test_text(self):
        """ Test text attribute """
        P = Passage(resource=xmlparser('<l n="8">Ibis <note>hello<a>b</a></note> ab excusso missus in astra <hi>sago.</hi> </l>'))
        # Without exclusion0
        self.assertEqual(P.text(), "Ibis hello b ab excusso missus in astra sago. ")
        # With Exclusion
        self.assertEqual(P.text(exclude=["note"]), "Ibis ab excusso missus in astra sago. ")

    def test_str(self):
        """ Test STR conversion of xml """
        P = Passage(resource=xmlparser('<l n="8">Ibis <note>hello<a>b</a></note> ab excusso missus in astra <hi>sago.</hi> </l>'))
        self.assertEqual(str(P), '<l n="8">Ibis <note>hello<a>b</a></note> ab excusso missus in astra <hi>sago.</hi> </l>')

    def test_xml(self):
        X = xmlparser('<l n="8">Ibis <note>hello<a>b</a></note> ab excusso missus in astra <hi>sago.</hi> </l>')
        P = Passage(resource=X)
        self.assertIs(X, P.xml)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unittest
import six

from io import open
from collections import defaultdict
from MyCapytain.resources.proto import inventory
from MyCapytain.common.reference import URN

class Resource(inventory.Resource):
    def parse(self, resource):
        return resource

class TIV(inventory.TextInventory):
    def parse(self, resource):
        self.textgroups = defaultdict(inventory.TextGroup)
        return resource

class TestRepoProto(unittest.TestCase):
    def test_resource_proto(self):
        with self.assertRaises(NotImplementedError):
            a = inventory.Resource(resource="hello")

    def test_implementation(self):
        a = Resource(resource="hello")
        self.assertEqual(a[0], a)
        self.assertIsNone(a[3.4])

    def test_equality(self):
        a = Resource(resource="hello")
        b = Resource(resource="hello2")
        c = Resource(resource=None)
        d = Resource(resource=None)

        self.assertNotEqual(a, b)
        self.assertNotEqual(a, 5)
        self.assertEqual((a == "i"), False)
        self.assertNotEqual(c, d)
        c.urn = "hello"
        d.urn = c.urn 
        self.assertEqual(c, d)
        c.resource = "AHAH"
        self.assertNotEqual(c, d)
        d.resource = "AHAH"
        self.assertEqual(c, d)

    def test_urn_access(self):
        # Limited to what's possible in proto...
        a = Resource(resource="hello")
        a.urn = URN("urn:cts:greekLit:tg")

        b = TIV(resource="hello")

        with six.assertRaisesRegex(self, ValueError, "Not valid urn"):
            a["urn:cts:greekLit"]

        self.assertEqual(a["urn:cts:greekLit:tg"], a)

        with six.assertRaisesRegex(self, ValueError, "Unrecognized urn at level"):
            b["urn:cts:greekLit:tg2"]
        
    def test_edit_trans(self):
        a = inventory.Edition()
        b = inventory.Translation()
        self.assertIsInstance(a, inventory.Text)
        self.assertIsInstance(b, inventory.Text)
        self.assertEqual(a.subtype, "Edition")
        self.assertEqual(b.subtype, "Translation")

    def test_write(self):
        a = Resource(resource="hello")
        with self.assertRaises(NotImplementedError):
            a.export()

    def test_str(self):
        a = Resource(resource="hello")
        with self.assertRaises(NotImplementedError):
            str(a)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unittest
import six

from io import open
from collections import defaultdict
from MyCapytain.resources.proto.text import *
import MyCapytain.common.reference
import MyCapytain.common.metadata

class TestProtoResource(unittest.TestCase):
    """ Test for resource, mother class of Text and Passage """
    def test_init(self):
        a = Resource()
        self.assertEqual(a.resource, None)

        a = Resource(resource=False)
        self.assertEqual(a.resource, False)
        
        a.resource = True
        self.assertEqual(a.resource, True)

    def test_urn(self):
        """ Test setters and getters for urn """
        a = Resource()
        # Should work with string
        a.urn = "urn:cts:latinLit:tg.wk.v" 
        self.assertEqual(isinstance(a.urn, MyCapytain.common.reference.URN), True)
        self.assertEqual(str(a.urn), "urn:cts:latinLit:tg.wk.v")

        # Test for URN
        a.urn = MyCapytain.common.reference.URN("urn:cts:latinLit:tg.wk.v2") 
        self.assertEqual(isinstance(a.urn, MyCapytain.common.reference.URN), True)
        self.assertEqual(str(a.urn), "urn:cts:latinLit:tg.wk.v2")

        # Test it fails if not basestring or URN
        with self.assertRaises(TypeError): 
            a.urn = 2

        # Test Resource setting works out as well
        b = Text(urn="urn:cts:latinLit:tg.wk.v")
        self.assertEqual(str(b.urn), "urn:cts:latinLit:tg.wk.v")


class TestProtoText(unittest.TestCase):
    """ Test the text prototype, mainly to ensure consistency """

    def test_init(self):
        """ Test init works correctly """
        a = Text()
        self.assertEqual(a.resource, None)

        a = Text(resource=False)
        self.assertEqual(a.resource, False)

        a.resource = True
        self.assertEqual(a.resource, True)

        #  Test with metadata
        a = Text()
        self.assertIsInstance(a.metadata, MyCapytain.common.metadata.Metadata)

        m = MyCapytain.common.metadata.Metadata(keys=["title", "author"])
        m["title"]["fre"] = "I am a metadata"
        a = Text(metadata=m)
        self.assertEqual(a.metadata["title"]["fre"], "I am a metadata")

    def test_proto_reff(self):
        """ Test that getValidReff function are not implemented """
        a = Text()

        # It should fail because not implemented
        with self.assertRaises(NotImplementedError):
            a.getValidReff()

        # It takes two parameters, should not issue TypeError
        # Consistency check
        with self.assertRaises(NotImplementedError):
            a.getValidReff(level=1, reference=["1"])


    def test_proto_passage(self):
        """ Test that getPassage function are not implemented but are consistent"""
        a = Text()

        # For getPassage, reference arg is required
        with self.assertRaises(TypeError): 
            a.getPassage()
        # It should fail because not implemented
        with self.assertRaises(NotImplementedError): 
            a.getPassage(reference=None)


    def test_get_label(self):
        """ Test that getLabel function are not implemented but are consistent"""
        a = Text()

        # It should fail because not implemented
        with self.assertRaises(NotImplementedError): 
            a.getLabel()

    def test_urn(self):
        """ Test setters and getters for urn """
        a = Text()
        # Should work with string
        a.urn = "urn:cts:latinLit:tg.wk.v" 
        self.assertEqual(isinstance(a.urn, MyCapytain.common.reference.URN), True)
        self.assertEqual(str(a.urn), "urn:cts:latinLit:tg.wk.v")

        # Test for URN
        a.urn = MyCapytain.common.reference.URN("urn:cts:latinLit:tg.wk.v2") 
        self.assertEqual(isinstance(a.urn, MyCapytain.common.reference.URN), True)
        self.assertEqual(str(a.urn), "urn:cts:latinLit:tg.wk.v2")

        # Test it fails if not basestring or URN
        with self.assertRaises(TypeError): 
            a.urn = 2

        # Test original setting works out as well
        b = Text(urn="urn:cts:latinLit:tg.wk.v")
        self.assertEqual(str(b.urn), "urn:cts:latinLit:tg.wk.v")


    def test_reffs(self):
        """ Test property reff, should fail because it supposes validReff is implemented """
        a = Text()

        # It should fail because not implemented
        with self.assertRaises(NotImplementedError):
            a.reffs

    def test_citation(self):
        """ Test citation property setter and getter """
        a = Text()
        a.citation = MyCapytain.common.reference.Citation(name="label")
        self.assertIsInstance(a.citation, MyCapytain.common.reference.Citation)

        #On init ?
        b = Text(citation=MyCapytain.common.reference.Citation(name="label"))
        self.assertEqual(a.citation.name, "label")

class TestProtoPassage(unittest.TestCase):
    """ Test the passage prototype, mainly to ensure consistency """
    def test_not_implemented(self):
        a = Passage()

        with self.assertRaises(NotImplementedError):
            a.next

        with self.assertRaises(NotImplementedError):
            a.prev

        with self.assertRaises(NotImplementedError):
            a.last

        with self.assertRaises(NotImplementedError):
            a.first

        with self.assertRaises(NotImplementedError):
            a.children


from __future__ import unicode_literals
from six import text_type as str

import MyCapytain.resources.proto.text
import MyCapytain.resources.texts.tei
import MyCapytain.resources.inventory
import MyCapytain.endpoints.proto
import MyCapytain.common.metadata
import MyCapytain.common.utils


class Text(MyCapytain.resources.proto.text.Text):
    """ Passage representing object prototype

    :param urn: A URN identifier
    :type urn: MyCapytain.common.reference.URN
    :param resource: An API endpoint
    :type resource: MyCapytain.endpoints.proto.CTS
    :param citation: Citation for children level
    :type citation: MyCapytain.resources.texts.tei.Citation
    :param id: Identifier of the subreference without URN informations
    :type id: List

    """

    DEFAULT_LANG = "eng"

    def __init__(self, urn, resource, citation=None, **kwargs):
        __doc__ = Text.__doc__
        super(Text, self).__init__(urn=urn, citation=citation, **kwargs)

        self._cRefPattern = None

        self.resource = resource

        if citation is not None:
            self.citation = citation

        if "metadata" in kwargs and isinstance(kwargs["metadata"], MyCapytain.common.metadata.Metadata):
            self.metadata = kwargs["metadata"]
        else:
            self.metadata = MyCapytain.common.metadata.Metadata([
                "groupname", "label", "title"
            ])

        self.passages = []

    def getValidReff(self, level=1, reference=None):
        """ Given a resource, Text will compute valid reffs

        :param level: Depth required. If not set, should retrieve first encountered level (1 based)
        :type level: Int
        :param reference: Subreference (optional)
        :type reference: Reference
        :rtype: List.basestring
        :returns: List of levels
        """
        if reference:
            urn = "{0}:{1}".format(self.urn, reference)
        else:
            urn = str(self.urn)

        if level == -1:
            level = len(self.citation)

        xml = self.resource.getValidReff(
            level=level,
            urn=urn
        )
        xml = MyCapytain.common.utils.xmlparser(xml)
        self.__parse_request(xml.xpath("//ti:request", namespaces=MyCapytain.common.utils.NS)[0])

        for ref in xml.xpath(
            "//ti:reply//ti:urn/text()",
            namespaces=MyCapytain.common.utils.NS
        ):
            self.passages.append(ref)

        return self.passages

    def getPassage(self, reference=None):
        """ Retrieve a passage and store it in the object

        :param reference: Reference of the passage
        :type reference: MyCapytain.common.reference.Reference or List of basestring
        :rtype: Passage
        :returns: Object representing the passage
        :raises: *TypeError* when reference is not a list or a Reference
        """
        if reference:
            urn = "{0}:{1}".format(self.urn, reference)
        else:
            urn = str(self.urn)

        response = MyCapytain.common.utils.xmlparser(self.resource.getPassage(urn=urn))

        self.__parse_request(response.xpath("//ti:request", namespaces=MyCapytain.common.utils.NS)[0])
        return Passage(urn=urn, resource=response, parent=self)

    def getPassagePlus(self, reference=None):
        """ Retrieve a passage and informations around it and store it in the object

        :param reference: Reference of the passage
        :type reference: MyCapytain.common.reference.Reference or List of basestring
        :rtype: Passage
        :returns: Object representing the passage
        :raises: *TypeError* when reference is not a list or a Reference
        """
        if reference:
            urn = "{0}:{1}".format(self.urn, reference)
        else:
            urn = str(self.urn)

        response = MyCapytain.common.utils.xmlparser(self.resource.getPassagePlus(urn=urn))

        self.__parse_request(response.xpath("//ti:reply/ti:label", namespaces=MyCapytain.common.utils.NS)[0])
        return Passage(urn=urn, resource=response, parent=self)

    def __parse_request(self, xml):
        """

        :param xml:
        :return:

        .. TODO: Finish self.citation parsing
        """
        for node in xml.xpath(".//ti:groupname", namespaces=MyCapytain.common.utils.NS):
            lang = node.get("xml:lang") or Text.DEFAULT_LANG
            self.metadata["groupname"][lang] = node.text

        for node in xml.xpath(".//ti:title", namespaces=MyCapytain.common.utils.NS):
            lang = node.get("xml:lang") or Text.DEFAULT_LANG
            self.metadata["title"][lang] = node.text

        for node in xml.xpath(".//ti:label", namespaces=MyCapytain.common.utils.NS):
            lang = node.get("xml:lang") or Text.DEFAULT_LANG
            self.metadata["label"][lang] = node.text

        # Need to code that p
        if self.citation is None:
            self.citation = MyCapytain.resources.inventory.Citation.ingest(
                xml,
                xpath=".//ti:citation[not(ancestor::ti:citation)]"
            )

    def getLabel(self):
        """ Retrieve metadata about the text

        :rtype: Metadata
        :returns: Dictionary with label informations
        """
        response = MyCapytain.common.utils.xmlparser(
            self.resource.getLabel(urn=str(self.urn))
        )

        self.__parse_request(response.xpath("//ti:reply/ti:label", namespaces=MyCapytain.common.utils.NS)[0])

        return self.metadata

    @property
    def reffs(self):
        """ Get all valid reffs for every part of the Text

        :rtype: MyCapytain.resources.texts.tei.Citation
        """
        if self.citation is None:
            reffs = [self.getValidReff()]
            return reffs + [
                reff for reffs in [self.getValidReff(level=i) for i in range(2, len(self.citation) + 1)] for reff in reffs
            ]
        else:
            return [
                reff for reffs in [self.getValidReff(level=i) for i in range(1, len(self.citation) + 1)] for reff in reffs
            ]


class Passage(MyCapytain.resources.texts.tei.Passage):

    def __init__(self, urn, resource, *args, **kwargs):
        super(Passage, self).__init__(resource=resource, *args, **kwargs)

        self.urn = urn

        # Could be set during parsing
        self._next = None
        self._prev = None
        self.__first = None
        self.__last = None

        self.__parse()

    @property
    def next(self):
        """ Following passage

        :rtype: Passage
        :returns: Following passage at same level
        """
        if self._next is not None:
            _next = self._next
        else:
            # Request the next urn
            _prev, _next = Passage.prevnext(
                self.parent.resource.getPrevNextUrn(urn=str(self.urn))
            )

        self.parent.resource.getPassage(urn=_next)

    @property
    def prev(self):
        """ Previous passage

        :rtype: Passage
        :returns: Previous passage at same level
        """
        if self._prev is not None:
           _prev = self._prev
        else:
            # Request the next urn
            _prev, _next = Passage.prevnext(
                self.parent.resource.getPrevNextUrn(urn=str(self.urn))
            )

        return self.parent.resource.getPassage(urn=_prev)

    def __parse(self):
        """ Given self.resource, split informations from the CTS API

        :return: None
        """
        self.resource = self.resource.xpath("//ti:passage/tei:TEI", namespaces=MyCapytain.common.utils.NS)[0]

        self._prev, self._next = Passage.prevnext(self.resource)

    @staticmethod
    def prevnext(resource):
        """ Parse a resource to get the prev and next urn

        :param resource: XML Resource
        :type resource: etree._Element
        :return: Tuple representing previous and next urn
        :rtype: (str, str)
        """
        _prev, _next = None, None
        resource = MyCapytain.common.utils.xmlparser(resource)
        prevnext = resource.xpath("//ti:prevnext", namespaces=MyCapytain.common.utils.NS)

        if len(prevnext) > 0:
            prevnext = prevnext[0]
            _next_xpath = prevnext.xpath("ti:next/ti:urn/text()", namespaces=MyCapytain.common.utils.NS)
            _prev_xpath = prevnext.xpath("ti:prev/ti:urn/text()", namespaces=MyCapytain.common.utils.NS)

            if len(_next_xpath):
               _next = _next_xpath[0]

            if len(_prev_xpath):
                _prev = _prev_xpath[0]

        return _prev, _next




import unittest

from MyCapytain.endpoints.proto import *


class TestEndpointsProto(unittest.TestCase):

    """ Testing prototypes endpoints """

    def setUp(self):
        self.ahab = Ahab("http://ahab.com")
        self.cts = CTS("http://ahab.com")

    def test_init_Ahab(self):
        """ Test that init register the Ahab endpoint url"""
        self.assertEqual(self.ahab.endpoint, "http://ahab.com")

    def test_raise_Ahab_proto(self):
        """ Tests that methods raises NotImplementedError """
        with self.assertRaises(NotImplementedError):
            self.ahab.search("query", "urn")

        with self.assertRaises(NotImplementedError):
            self.ahab.permalink("urn")

    def test_raise_Ahab_search_arguments(self):
        """ Tests that methods have consistent arguments"""
        with self.assertRaises(NotImplementedError):
            self.ahab.search(
                query="query", urn="urn", start=1, limit=5, format="json")

    def test_raise_Ahab_permalink_arguments(self):
        """ Tests that methods have consistent arguments"""
        with self.assertRaises(NotImplementedError):
            self.ahab.permalink(urn="urn", format="json")

    def test_init_CTS(self):
        """ Test that init register the CTS endpoint url"""
        self.assertEqual(self.ahab.endpoint, "http://ahab.com")

    def test_raise_CTS_proto(self):
        """ Tests that methods raises NotImplementedError """
        with self.assertRaises(NotImplementedError):
            self.cts.getCapabilities("inventory")

        with self.assertRaises(NotImplementedError):
            self.cts.getValidReff("urn", "inventory")

        with self.assertRaises(NotImplementedError):
            self.cts.getPassage("urn", "inventory")

        with self.assertRaises(NotImplementedError):
            self.cts.getPassagePlus("urn", "inventory")

        with self.assertRaises(NotImplementedError):
            self.cts.getFirstUrn("urn", "inventory")

        with self.assertRaises(NotImplementedError):
            self.cts.getPrevNextUrn("urn", "inventory")

        with self.assertRaises(NotImplementedError):
            self.cts.getLabel("urn", "inventory")

    def test_raise_CTS_getCapabilities_arguments(self):
        """ Tests that methods getCapabilities have consistent arguments"""
        with self.assertRaises(NotImplementedError):
            self.cts.getCapabilities(inventory="inventory")

    def test_raise_CTS_getValidReff_arguments(self):
        """ Tests that methods getValidReff have consistent arguments"""
        with self.assertRaises(NotImplementedError):
            self.cts.getValidReff(urn="urn", inventory="inventory", level=1)

    def test_raise_CTS_getPassage_arguments(self):
        """ Tests that methods getPassage have consistent arguments"""
        with self.assertRaises(NotImplementedError):
            self.cts.getPassage(urn="urn", inventory="inventory", context=1)

    def test_raise_CTS_getPassagePlus_arguments(self):
        """ Tests that methods getPassagePlus have consistent arguments"""
        with self.assertRaises(NotImplementedError):
            self.cts.getPassagePlus(urn="urn", inventory="inventory", context=1)

    def test_raise_CTS_getFirstUrn_arguments(self):
        """ Tests that methods getFirstUrn have consistent arguments"""
        with self.assertRaises(NotImplementedError):
            self.cts.getFirstUrn(urn="urn", inventory="inventory")

    def test_raise_CTS_getPrevNextUrn_arguments(self):
        """ Tests that methods getPrevNextUrn have consistent arguments"""
        with self.assertRaises(NotImplementedError):
            self.cts.getPrevNextUrn(urn="urn", inventory="inventory")

    def test_raise_CTS_getLabel_arguments(self):
        """ Tests that methods getLabel have consistent arguments"""
        with self.assertRaises(NotImplementedError):
            self.cts.getLabel(urn="urn", inventory="inventory")


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unittest
from six import text_type as str
from io import open
import xmlunittest

import MyCapytain.resources.texts.local
import MyCapytain.resources.texts.tei
import MyCapytain.common.reference
        

class TestLocalXMLTextImplementation(unittest.TestCase, xmlunittest.XmlTestMixin):

    """ Test XML Implementation of resources found in local file """

    def setUp(self):
        self.text = open("tests/testing_data/texts/sample.xml", "rb")
        self.TEI = MyCapytain.resources.texts.local.Text(resource=self.text)

    def tearDown(self):
        self.text.close()

    def testFindCitation(self):
        self.assertEqual(
            str(self.TEI.citation),
            '<tei:cRefPattern n="book" matchPattern="(\\w+)" replacementPattern="#xpath(/tei:TEI/tei:text/tei:body/tei:div/tei:div[@n=\\\'$1\\\'])"><tei:p>This pointer pattern extracts book</tei:p></tei:cRefPattern>'
        )
        self.assertEqual(
            str(self.TEI.citation.child),
            '<tei:cRefPattern n="poem" matchPattern="(\\w+)\.(\\w+)" replacementPattern="#xpath(/tei:TEI/tei:text/tei:body/tei:div/tei:div[@n=\\\'$1\\\']/tei:div[@n=\\\'$2\\\'])"><tei:p>This pointer pattern extracts poem</tei:p></tei:cRefPattern>'
        )
        self.assertEqual(
            str(self.TEI.citation.child.child),
            '<tei:cRefPattern n="line" matchPattern="(\\w+)\.(\\w+)\.(\\w+)" replacementPattern="#xpath(/tei:TEI/tei:text/tei:body/tei:div/tei:div[@n=\\\'$1\\\']/tei:div[@n=\\\'$2\\\']/tei:l[@n=\\\'$3\\\'])"><tei:p>This pointer pattern extracts line</tei:p></tei:cRefPattern>'
        )

        self.assertEqual(len(self.TEI.citation), 3)

    def testCitationSetters(self):
        d = MyCapytain.resources.texts.tei.Citation()
        c = MyCapytain.common.reference.Citation(name="ahah", refsDecl="/tei:TEI/tei:text/tei:body/tei:div/tei:div[@n='$1']", child=None)
        b = MyCapytain.resources.texts.tei.Citation()
        a = MyCapytain.resources.texts.local.Text(citation=b)
        """ Test original setting """
        self.assertIs(a.citation, b)
        """ Test simple replacement """
        a.citation = d
        self.assertIs(a.citation, d)
        """ Test conversion """
        a.citation = c
        self.assertEqual(a.citation.name, "ahah")
        self.assertEqual(a.citation.child, None)
        self.assertEqual(a.citation.refsDecl, "/tei:TEI/tei:text/tei:body/tei:div/tei:div[@n='$1']")
        self.assertEqual(a.citation.scope, "/tei:TEI/tei:text/tei:body/tei:div")
        self.assertEqual(a.citation.xpath, "/tei:div[@n='?']")

    def testFindCitation(self):
        # Test level 1
        self.assertEqual(self.TEI.getValidReff(), ["1", "2"])
        # Test level 2
        self.assertEqual(self.TEI.getValidReff(level=2)[0], "1.pr")
        # Test level 3
        self.assertEqual(self.TEI.getValidReff(level=3)[0], "1.pr.1")
        self.assertEqual(self.TEI.getValidReff(level=3)[-1], "2.40.8")

        # Test with reference and level
        self.assertEqual(self.TEI.getValidReff(reference=MyCapytain.common.reference.Reference("2.1"),level=3)[1], "2.1.2")
        self.assertEqual(self.TEI.getValidReff(reference=MyCapytain.common.reference.Reference("2.1"),level=3)[-1], "2.1.12")

        # Test with reference and level autocorrected because too small
        self.assertEqual(self.TEI.getValidReff(reference=MyCapytain.common.reference.Reference("2.1"),level=0)[-1], "2.1.12")
        self.assertEqual(self.TEI.getValidReff(reference=MyCapytain.common.reference.Reference("2.1"),level=2)[-1], "2.1.12")

        # Test when already too deep
        self.assertEqual(self.TEI.getValidReff(reference=MyCapytain.common.reference.Reference("2.1.1"),level=3), [])

        # Test wrong citation
        with self.assertRaises(KeyError): 
            self.assertEqual(self.TEI.getValidReff(reference=MyCapytain.common.reference.Reference("2.hellno"),level=3), [])

    def test_wrong_main_scope(self):
        with open("tests/testing_data/texts/sample2.xml", "rb") as file:
            with self.assertRaises(MyCapytain.resources.texts.local.RefsDeclError):
                text = MyCapytain.resources.texts.local.Text(resource=file)

    def test_reffs(self):
        """ Check that every level is returned trough reffs property """
        self.assertEqual(("1" in self.TEI.reffs), True)
        self.assertEqual(("1.pr" in self.TEI.reffs), True)
        self.assertEqual(("2.40.8" in self.TEI.reffs), True)

    def test_urn(self):
        """ Test setters and getters for urn """

        # Should work with string
        self.TEI.urn = "urn:cts:latinLit:tg.wk.v" 
        self.assertEqual(isinstance(self.TEI.urn, MyCapytain.common.reference.URN), True)
        self.assertEqual(str(self.TEI.urn), "urn:cts:latinLit:tg.wk.v")

        # Test for URN
        self.TEI.urn = MyCapytain.common.reference.URN("urn:cts:latinLit:tg.wk.v2") 
        self.assertEqual(isinstance(self.TEI.urn, MyCapytain.common.reference.URN), True)
        self.assertEqual(str(self.TEI.urn), "urn:cts:latinLit:tg.wk.v2")

        # Test it fails if not basestring or URN
        with self.assertRaises(TypeError): 
            self.TEI.urn = 2

    def test_get_passage(self):
        a = self.TEI.getPassage(["1", "pr", "2"])
        self.assertEqual(a.text(), "tum, ut de illis queri non possit quisquis de se bene ")
        # With reference
        a = self.TEI.getPassage(MyCapytain.common.reference.Reference("2.5.5"))
        self.assertEqual(a.text(), "Saepe domi non es, cum sis quoque, saepe negaris: ")

    def test_get_passage_plus(self):
        """ Test GetPassage Plus """
        # No label in local 
        a = self.TEI.getPassagePlus(["1", "pr", "2"])

        self.assertEqual(a.prev, ["1", "pr", "1"])
        self.assertEqual(a.next, ["1", "pr", "3"])
        self.assertEqual(a.passage.text(), "tum, ut de illis queri non possit quisquis de se bene ")


class TestLocalXMLPassageImplementation(unittest.TestCase, xmlunittest.XmlTestMixin):
    """ Test passage implementation """

    def setUp(self):
        self.URN = MyCapytain.common.reference.URN("urn:cts:latinLit:phi1294.phi002.perseus-lat2")
        self.URN_2 = MyCapytain.common.reference.URN("urn:cts:latinLit:phi1294.phi002.perseus-lat3")
        self.text = open("tests/testing_data/texts/sample.xml", "rb")
        self.TEI = MyCapytain.resources.texts.local.Text(resource=self.text)

    def tearDown(self):
        self.text.close()

    def test_urn(self):
        """ Test URN and ids getters/setters """

        a = MyCapytain.resources.texts.local.Passage()
 
        #~Test simple set up
        a.urn = self.URN
        self.assertEqual(str(a.urn), "urn:cts:latinLit:phi1294.phi002.perseus-lat2")
        # Test update on ID update
        a.id = ["1", "pr", "1"]
        self.assertEqual(str(a.urn), "urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.pr.1")
        # Should keep the ID if URN changes
        a.urn = self.URN_2
        self.assertEqual(str(a.urn), "urn:cts:latinLit:phi1294.phi002.perseus-lat3:1.pr.1")
        # Test init
        a = MyCapytain.resources.texts.local.Passage(urn=self.URN)
        self.assertEqual(str(a.urn), "urn:cts:latinLit:phi1294.phi002.perseus-lat2")
        # Test init with id and URN
        a = MyCapytain.resources.texts.local.Passage(urn=self.URN, id=["1", "pr", "1"])
        self.assertEqual(str(a.urn), "urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.pr.1")
        # Should raise error if not URN for consistency
        with self.assertRaises(TypeError):
            a.urn = 1
        # Should work with plain text
        a = MyCapytain.resources.texts.local.Passage()
        a.urn = "urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.pr.1"
        self.assertEqual(str(a.urn), "urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.pr.1")
        self.assertEqual(a.id, ["1", "pr", "1"])
        # This should affect id !
        # Test Id value on init
        a = MyCapytain.resources.texts.local.Passage(id=["1", "pr", "1"])
        self.assertEqual(a.id, ["1", "pr", "1"])

    def test_next(self):
        """ Test next property """
        # Normal passage checking
        p = self.TEI.getPassage(["1", "pr", "1"])
        self.assertEqual(p.next.id, ["1", "pr", "2"])

        # End of lowest level passage checking but not end of parent level
        p = self.TEI.getPassage(["1", "pr", "22"])
        self.assertEqual(p.next.id, ["1", "1", "1"])

        # End of lowest level passage and end of parent level
        p = self.TEI.getPassage(["1", "39", "8"])
        self.assertEqual(p.next.id, ["2", "pr", "sa"])

        # Last line should always be None
        p = self.TEI.getPassage(["2", "40", "8"])
        self.assertIsNone(p.next)
        p = self.TEI.getPassage(["2", "40"])
        self.assertIsNone(p.next)
        p = self.TEI.getPassage(["2"])
        self.assertIsNone(p.next)

    def test_children(self):
        """ Test children property """
        # Normal children checking
        with open("tests/testing_data/texts/sample.xml", "rb") as text:
            self.TEI = MyCapytain.resources.texts.local.Text(resource=text)

            p = self.TEI.getPassage(["1", "pr"])
            self.assertEqual(p.children["1.pr.1"].id, ["1", "pr", "1"])

            p = self.TEI.getPassage(["1", "pr", "1"])
            self.assertEqual(len(p.children), 0)

    def test_first(self):
        """ Test first property """
        # Test when there is one
        p = self.TEI.getPassage(["1", "pr"])
        self.assertEqual(p.first.id, ["1", "pr", "1"])
        # #And failing when no first
        p = self.TEI.getPassage(["1", "pr", "1"])
        self.assertEqual(p.first, None)

    def test_last(self):
        """ Test last property """
        # Test when there is one
        p = self.TEI.getPassage(["1", "pr"])
        self.assertEqual(p.last.id, ["1", "pr", "22"])
        # #And failing when no last
        p = self.TEI.getPassage(["1", "pr", "1"])
        self.assertEqual(p.last, None)

    def test_prev(self):
        """ Test prev property """
        # Normal passage checking
        p = self.TEI.getPassage(["2", "40", "8"])
        self.assertEqual(p.prev.id, ["2", "40", "7"])
        p = self.TEI.getPassage(["2", "40"])
        self.assertEqual(p.prev.id, ["2", "39"])
        p = self.TEI.getPassage(["2"])
        self.assertEqual(p.prev.id, ["1"])

        # test failing passage
        p = self.TEI.getPassage(["1", "pr", "1"])
        self.assertEqual(p.prev, None)
        p = self.TEI.getPassage(["1", "pr"])
        self.assertEqual(p.prev, None)
        p = self.TEI.getPassage(["1"])
        self.assertEqual(p.prev, None)

        # First child should get to parent's prev last child
        p = self.TEI.getPassage(["1", "1", "1"])
        self.assertEqual(p.prev.id, ["1", "pr", "22"])

        # Beginning of lowest level passage and beginning of parent level
        p = self.TEI.getPassage(["2", "pr", "sa"])
        self.assertEqual(p.prev.id, ["1", "39", "8"])


# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.resources.xml


Local files handler for CTS

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""

from collections import OrderedDict, defaultdict
from past.builtins import basestring

from MyCapytain.common.utils import xmlparser, NS
from MyCapytain.common.reference import URN, Citation, Reference
from MyCapytain.resources.proto import text
import MyCapytain.resources.texts.tei


class RefsDeclError(Exception):
    """ Error issued when an the refsDecl does not succeed in xpath (no results)
    """
    pass


class Text(text.Text):
    """ Implementation of CTS tools for local files

    :param urn: A URN identifier
    :type urn: MyCapytain.common.reference.URN
    :param resource: A resource
    :type resource: lxml.etree._Element
    :param citation: Highest Citation level
    :type citation: MyCapytain.common.reference.Citation

    :ivar resource: lxml
    """

    def __init__(self, urn=None, citation=None, resource=None):
        self._passages = OrderedDict() # Represents real full passages / reffs informations. Only way to set it up is getValidReff without passage ?
        self._orphan = defaultdict(Reference) # Represents passage we got without asking for all. Storing convenience ?

        self._cRefPattern = MyCapytain.resources.texts.tei.Citation()
        self.resource = None
        self.xml = None
        self._URN = None

        if citation is not None:
            self.citation = citation
        if resource is not None:
            self.resource = resource
            self.xml = xmlparser(resource)

            self.__findCRefPattern(self.xml)

            try:
                xml = self.xml.xpath(self.citation.scope, namespaces=NS)[0]
            except IndexError:
                msg = "Main citation scope does not result in any result ({0})".format(self.citation.scope)
                raise RefsDeclError(msg)
            except Exception as E:
                raise E

            self._passages = Passage(resource=xml, citation=self.citation, urn=self.urn, id=None)

    def __findCRefPattern(self, xml):
        self.citation = MyCapytain.resources.texts.tei.Citation.ingest(
            resource=xml.xpath("//tei:refsDecl[@n='CTS']", namespaces=MyCapytain.common.utils.NS),
            xpath=".//tei:cRefPattern"
        )

    @property
    def citation(self):
        """ Get the lowest cRefPattern in the hierarchy

        :rtype: MyCapytain.resources.texts.tei.Citation
        """
        return self._cRefPattern

    @citation.setter
    def citation(self, value):
        """ Set the cRefPattern

        :param value: Citation to be saved
        :type value:  MyCapytain.resources.texts.tei.Citation or Citation
        :raises: TypeError when value is not a TEI Citation or a Citation
        """
        if isinstance(value, MyCapytain.resources.texts.tei.Citation):
            self._cRefPattern = value
        elif isinstance(value, Citation):
            # .. todo:: Should support conversion between Citation...
            self._cRefPattern = MyCapytain.resources.texts.tei.Citation(
                name=value.name,
                refsDecl=value.refsDecl,
                child=value.child
            )

    def getPassage(self, reference):
        """ Finds a passage in the current text

        :param reference: Identifier of the subreference / passages
        :type reference: List, MyCapytain.common.reference.Reference
        :rtype: Passage
        :returns: Asked passage
        """
        if isinstance(reference, MyCapytain.common.reference.Reference):
            reference = reference["start_list"]

        reference = [".".join(reference[:i]) for i in range(1, len(reference) + 1 )]
        passages = [self._passages]
        while len(reference) > 0:
            passages = [passage for sublist in [p.get(reference[0]) for p in passages] for passage in sublist]
            reference.pop(0)

        return passages[0]

    def getPassagePlus(self, reference):
        """ Finds a passage in the current text with its previous and following node

        :param reference: Identifier of the subreference / passages
        :type reference: List, MyCapytain.common.reference.Reference
        :rtype: text.PassagePlus
        :returns: Asked passage with metainformations
        """
        P = self.getPassage(reference=reference)
        return text.PassagePlus(P, P.prev.id, P.next.id)

    def getValidReff(self, level=1, reference=None):
        """ Retrieve valid passages directly

        :param level: Depth required. If not set, should retrieve first encountered level (1 based)
        :type level: Int
        :param reference: Subreference (optional)
        :type reference: Reference
        :rtype: List.basestring
        :returns: List of levels

        .. note:: GetValidReff works for now as a loop using Passage, subinstances of Text, to retrieve the valid informations. Maybe something is more powerfull ?
        """

        if reference is not None:
            start = len(reference[2])
            nodes = [".".join(reference[2][0:i+1]) for i in range(0, start)] + [None]
            if level <= start:
                level = start + 1
        else:
            nodes = [None for i in range(0, level)]

        passages = [self._passages] # For consistency
        while len(nodes) >= 1:
            passages = [passage for sublist in [p.get(nodes[0]) for p in passages] for passage in sublist]
            nodes.pop(0)

        return [".".join(passage.id) for passage in passages]


class Passage(MyCapytain.resources.texts.tei.Passage):
    """ Passage representing object
    
    :param urn: A URN identifier
    :type urn: MyCapytain.common.reference.URN
    :param resource: A resource
    :type resource: lxml.etree._Element
    :param parent: Parent of the current passage
    :type parent: MyCapytain.resources.texts.tei.Passage
    :param citation: Citation for children level
    :type citation: MyCapytain.resources.texts.tei.Citation
    :param id: Identifier of the subreference without URN informations
    :type id: List

    .. note:: *id* is used in to identify the current passage in case the URN is unknown
    """

    def __init__(self, urn=None, resource=None, parent=None, citation=None, id=None):
        super(Passage, self).__init__(urn=urn, resource=resource, parent=parent)

        self.__next = False
        self.__prev = False

        self.citation = None
        if isinstance(citation, Citation):
            self.citation = citation

        self.__id = []

        if id is not None:
            self.id = id

        self.__children = OrderedDict()
        self.__parsed = False

    @property
    def id(self):
        """ Id represents the passage subreference as a list of basestring

        :rtype: list
        :returns: Representation of the passage subreference as a list
        """
        return self.__id

    @id.setter
    def id(self, value):
        """ Set up ID property

        :param value: Representation of the passage subreference as a list
        :type value: list

        .. note:: `Passage.id = [..]` will update automatically the URN property as well if correct
        """
        if isinstance(value, (list, tuple)):
            self.__id = value
            self.__updateURN()

    @property
    def urn(self):
        """ URN Identifier of the object

        :rtype: MyCapytain.common.reference.URN

        """
        return self._URN

    @urn.setter
    def urn(self, value):
        """ Set the urn

        :param value: URN to be saved
        :type value:  MyCapytain.common.reference.URN
        :raises: *TypeError* when the value is not URN compatible

        .. note:: `Passage.URN = ...` will update automatically the id property if Passage is set

        """
        a = self._URN
        if isinstance(value, basestring):
            value = MyCapytain.common.reference.URN(value)
        elif not isinstance(value, MyCapytain.common.reference.URN):
            raise TypeError()
        if str(a) != str(value):
            self._URN = value
            self.__updateURN()

    def __updateURN(self):
        """ Private method allowing for update of self.id or self.urn
        """
        if self.id is not None and len(self.id) > 0 and isinstance(self.urn, URN):
            self.urn = URN(self.urn["text"] + ":" + ".".join(self.id))
        elif self._URN is not None \
            and self._URN["reference"] is not None \
            and len(self._URN["reference"][2]) > 0 \
            and self.id != self._URN["reference"][2]:

            self.__id = self._URN["reference"][2]

    def get(self, key=None):
        """ Get a child or multiple children

        :param key: String identifying a passage
        :type key: basestring or int 
        :raises KeyError: When key identifies a child unknown to this passage
        :rtype: List.Passage
        :returns: List of passage identified by key. If key is None, returns all children
        
        .. note:: Call time depends on parsing status. If the passage was never parsed, then on first call citation is used to find children
        """
        if len(self.__children) == 0 and self.__parsed is False:
            self.__parse()

        if key is None:
            return [self.__children[key] for key in self.__children]
        elif isinstance(key, int):
            keys = list(self.__children.copy().keys())
            return [self.__children[keys[key]]]
        elif key not in self.__children:
            raise KeyError()
        else:
            return [self.__children[key]]

    def __parse(self):
        """ Private method for parsing children
        """
        if self.citation is None:
            self.__parsed = True
            return []
        elements = self.resource.xpath("."+self.citation.fill(passage=None, xpath=True), namespaces=NS)
        for element in elements:
            n = self.id + [element.get("n")]
            self.__children[".".join(n)] = Passage(
                resource=element,
                citation=self.citation.child,
                id=n,
                urn=self.urn,
                parent=self
            )
        self.__parsed = True

    @property
    def first(self):
        """ First child of current Passage 
        
        :rtype: None or Passage
        :returns: None if current Passage has no children,  first child passage if available
        """
        try:
            return self.get(0)[0]
        except:
            return None

    @property
    def last(self):
        """ Last child of current Passage 
        
        :rtype: None or Passage
        :returns: None if current Passage has no children, last child passage if available
        """
        try:
            return self.get(-1)[0]
        except :
            return None

    @property
    def children(self):
        """ Children of the passage

        :rtype: OrderedDict
        :returns: Dictionary of chidren, where key are subreferences
        """
        if len(self.__children) == 0 and self.__parsed is False:
            self.__parse()

        return self.__children
    
    @property
    def next(self):
        """ Next passage 

        :rtype: Passage
        :returns: Next passage at same level
        """ 
        if self.__next is False:

            if self.parent is None: # When top of hierarchy is access, should return None
                self.__next = None
                return None

            keys = list(self.parent.children.copy().keys())
            current = keys.index(".".join(self.id))
            if len(keys) - 1 > current:
                self.__next = self.parent.get(keys[current + 1])[0]
            else:
                n = self.parent.next
                if n is None:
                    self.__next = None
                else:
                    self.__next = n.first

        return self.__next

    @property
    def prev(self):
        """ Previous passage 

        :rtype: Passage
        :returns: Previous passage at same level
        """ 
        if self.__prev is False:

            if self.parent is None: # When top of hierarchy is access, should return None
                self.__prev = None
                return None

            keys = list(self.parent.children.copy().keys())
            current = keys.index(".".join(self.id))
            if current > 0:
                self.__prev = self.parent.get(keys[current - 1])[0]
            else:
                n = self.parent.prev
                if n is None:
                    self.__prev = None
                else:
                    self.__prev = n.last

        return self.__prev


# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.common.reference
   :synopsis: Common useful tools and constants

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""
from __future__ import unicode_literals


from lxml import etree
from io import IOBase, StringIO
from past.builtins import basestring
import re

__strip = re.compile("([ ]{2,})+")


def normalize(string):
    """ Remove double-or-more spaces in a string

    :param string: A string to change
    :type string: basestring
    :rtype: Basestring
    :returns: Clean string
    """
    return __strip.sub(" ", string)

#: Dictionary of namespace that can be useful
NS = {
    "tei": "http://www.tei-c.org/ns/1.0",
    "ahab": "http://localhost.local",
    "ti": "http://chs.harvard.edu/xmlns/cts",
    "xml": "http://www.w3.org/XML/1998/namespace"
}


def xmlparser(xml):
    """ Parse xml 

    :param xml: XML element
    :type xml: basestring, lxml.etree._Element
    :rtype: lxml.etree._Element
    :returns: An element object
    :raises: TypeError if element is not in accepted type

    """
    doclose = None
    if isinstance(xml, etree._Element):
        return xml
    elif isinstance(xml, IOBase):
        pass
    elif isinstance(xml, basestring):
        xml = StringIO(xml)
        doclose = True
    else:
        raise TypeError("Unsupported type of resource")
    parsed = etree.parse(xml).getroot()
    if doclose:
        xml.close()
    return parsed




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unittest
from io import open
import lxml.etree as etree
import lxml.objectify
import xmlunittest
from copy import deepcopy
from six import text_type as str

from MyCapytain.resources.inventory import *
import MyCapytain.resources.proto.text


def compareSTR(one, other):
    return (one.replace("\n", ""), other.replace("\n", ""))

def compareXML(one, other):
    return (etree.tostring(one, encoding=str).replace("\n", ""), other.replace("\n", ""))

class TestXMLImplementation(unittest.TestCase, xmlunittest.XmlTestMixin):

    """ Test XML Implementation of resources Endpoint request making """

    def setUp(self):
        self.getCapabilities = open("tests/testing_data/cts/getCapabilities.xml", "r")
        self.ed = """<ti:edition urn='urn:cts:latinLit:phi1294.phi002.perseus-lat2' workUrn='urn:cts:latinLit:phi1294.phi002' xmlns:ti='http://chs.harvard.edu/xmlns/cts'>
<ti:label xml:lang='eng'>Epigrammata Label</ti:label>
<ti:label xml:lang='fre'>Epigrammes Label</ti:label>
<ti:description xml:lang='eng'>W. Heraeus</ti:description>
<ti:description xml:lang='fre'>G. Heraeus</ti:description>
<ti:online></ti:online>
</ti:edition>""".replace("\n", "")

        self.tr = """<ti:translation xml:lang='eng' urn='urn:cts:latinLit:phi1294.phi002.perseus-eng2' workUrn='urn:cts:latinLit:phi1294.phi002' xmlns:ti='http://chs.harvard.edu/xmlns/cts'>
<ti:label xml:lang='eng'>Epigrammata</ti:label>
<ti:description xml:lang='eng'>M. Valerii Martialis Epigrammaton libri / recognovit W. Heraeus</ti:description>
<ti:online></ti:online>
</ti:translation>""".replace("\n", "")

        self.wk = """<ti:work urn='urn:cts:latinLit:phi1294.phi002' groupUrn='urn:cts:latinLit:phi1294' xmlns:ti='http://chs.harvard.edu/xmlns/cts'>
<ti:title xml:lang='eng'>Epigrammata</ti:title>
<ti:title xml:lang='fre'>Epigrammes</ti:title>""" + self.tr + self.ed + """</ti:work>""".replace("\n", "")

        self.tg = """<ti:textgroup urn='urn:cts:latinLit:phi1294' xmlns:ti='http://chs.harvard.edu/xmlns/cts'>
<ti:groupname xml:lang='eng'>Martial</ti:groupname>
<ti:groupname xml:lang='lat'>Martialis</ti:groupname>""" + self.wk + """</ti:textgroup>""".replace("\n", "")

        self.t = """<ti:TextInventory tiid='annotsrc' xmlns:ti='http://chs.harvard.edu/xmlns/cts'>""" + self.tg + """</ti:TextInventory>""".replace("\n", "").strip("\n")


    def tearDown(self):
        self.getCapabilities.close()

    def test_xml_TextInventoryParsing(self):
        """ Tests TextInventory parses without errors """
        TI = TextInventory(resource=self.getCapabilities, id="TestInv")
        self.assertGreater(len(TI.textgroups), 0)

    def test_xml_TextInventory_GetItem(self):
        """ Test access through getItem obj[urn] """
        TI = TextInventory(resource=self.getCapabilities, id="TestInv")
        self.assertIsInstance(TI["urn:cts:latinLit:phi1294"], TextGroup)
        self.assertIsInstance(TI["urn:cts:latinLit:phi1294.phi002"], Work)
        self.assertEqual(str(TI["urn:cts:latinLit:phi1294.phi002"].urn), "urn:cts:latinLit:phi1294.phi002")
        self.assertIsInstance(TI["urn:cts:latinLit:phi1294.phi002.perseus-lat2"], Text)
        self.assertEqual(str(TI["urn:cts:latinLit:phi1294.phi002.perseus-lat2"].urn), "urn:cts:latinLit:phi1294.phi002.perseus-lat2")

    def test_xml_Work_GetItem(self):
        """ Test access through getItem obj[urn] """
        TI = TextInventory(resource=self.getCapabilities, id="TestInv")
        tg = TI["urn:cts:latinLit:phi1294"]
        self.assertIsInstance(tg["urn:cts:latinLit:phi1294.phi002"], Work)
        self.assertEqual(str(tg["urn:cts:latinLit:phi1294.phi002"].urn), "urn:cts:latinLit:phi1294.phi002")
        self.assertIsInstance(tg["urn:cts:latinLit:phi1294.phi002.perseus-lat2"], Text)
        self.assertEqual(str(tg["urn:cts:latinLit:phi1294.phi002.perseus-lat2"].urn), "urn:cts:latinLit:phi1294.phi002.perseus-lat2")

    def test_xml_work_getLang(self):
        """ Test access to translation """
        xml = """
            <ti:work xmlns:ti="http://chs.harvard.edu/xmlns/cts" urn="urn:cts:latinLit:phi1294.phi002">
                <ti:title xml:lang="eng">Epigrammata</ti:title>
                <ti:edition workUrn="urn:cts:latinLit:phi1294.phi002" urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2">
                </ti:edition>
                <ti:translation workUrn="urn:cts:latinLit:phi1294.phi002" urn="urn:cts:latinLit:phi1294.phi002.perseus-eng2" xml:lang="eng">
                </ti:translation>
                <ti:translation workUrn="urn:cts:latinLit:phi1294.phi002" urn="urn:cts:latinLit:phi1294.phi002.perseus-eng3" xml:lang="eng">
                </ti:translation>
                <ti:translation workUrn="urn:cts:latinLit:phi1294.phi002" urn="urn:cts:latinLit:phi1294.phi002.perseus-fre1" xml:lang="fre">
                </ti:translation>
            </ti:work>
        """
        W = Work(resource=xml, urn="urn:cts:latinLit:phi1294.phi002")
        self.assertEqual(len(W.getLang("eng")), 2)
        self.assertEqual(len(W.getLang()), 3)

    def test_xml_Text_others(self):
        """ Test access to translation """
        xml = """
            <ti:work xmlns:ti="http://chs.harvard.edu/xmlns/cts" urn="urn:cts:latinLit:phi1294.phi002">
                <ti:title xml:lang="eng">Epigrammata</ti:title>
                <ti:edition workUrn="urn:cts:latinLit:phi1294.phi002" urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2">
                </ti:edition>
                <ti:translation workUrn="urn:cts:latinLit:phi1294.phi002" urn="urn:cts:latinLit:phi1294.phi002.perseus-eng2" xml:lang="eng">
                </ti:translation>
                <ti:translation workUrn="urn:cts:latinLit:phi1294.phi002" urn="urn:cts:latinLit:phi1294.phi002.perseus-eng3" xml:lang="eng">
                </ti:translation>
                <ti:translation workUrn="urn:cts:latinLit:phi1294.phi002" urn="urn:cts:latinLit:phi1294.phi002.perseus-fre1" xml:lang="fre">
                </ti:translation>
            </ti:work>
        """
        W = Work(resource=xml, urn="urn:cts:latinLit:phi1294.phi002")
        E = W["urn:cts:latinLit:phi1294.phi002.perseus-lat2"]
        T = W["urn:cts:latinLit:phi1294.phi002.perseus-fre1"]

        self.assertEqual(E.translations("fre"), [T])
        self.assertEqual(T.editions(), [E])

    def test_get_parent(self):
        TI = TextInventory(resource=self.getCapabilities, id="TestInv")
        tg = TI["urn:cts:latinLit:phi1294"]
        wk = TI["urn:cts:latinLit:phi1294.phi002"]
        tx = TI["urn:cts:latinLit:phi1294.phi002.perseus-lat2"]
        self.assertEqual(tg[1], TI)
        self.assertEqual(wk[1], tg)
        self.assertEqual(wk[2], TI)
        self.assertEqual(tx[1], wk)
        self.assertEqual(tx[2], tg)
        self.assertEqual(tx[3], TI)

    def test_translation(self):
        TI = TextInventory(resource=self.getCapabilities, id="TestInv")
        tr = TI["urn:cts:latinLit:phi1294.phi002.perseus-eng2"]
        self.assertIsInstance(tr, Text)
        self.assertEqual(tr.subtype, "Translation")

    def test_parse_string(self):
        TI = TextInventory(
            id="TestInv",
            resource="""
<ti:TextInventory xmlns:ti="http://chs.harvard.edu/xmlns/cts" tiid="thibault3">
    <ti:textgroup urn="urn:cts:greekLit:tlg0003">
        <ti:groupname xml:lang="en">Thucydides</ti:groupname>
        <ti:groupname xml:lang="eng">Thucydides</ti:groupname>
        <ti:groupname>Thucydides</ti:groupname>
        <ti:work urn="urn:cts:greekLit:tlg0003.tlg001" xml:lang="grc">
            <ti:title xml:lang="en">The Peloponnesian War</ti:title>
            <ti:title xml:lang="eng">The Peloponnesian War</ti:title>
        </ti:work>
    </ti:textgroup>
</ti:TextInventory>
            """
        )
        self.assertEqual(str(TI["urn:cts:greekLit:tlg0003.tlg001"].urn), "urn:cts:greekLit:tlg0003.tlg001")

    def test_parse_error(self):
        with self.assertRaises(TypeError):
            TI = TextInventory(
                id="TestInv",
                resource=5
            )

    def test_Inventory_metadata(self):
        """ Tests TextInventory parses without errors """
        TI = TextInventory(resource=self.getCapabilities, id="annotsrc")
        self.assertEqual(TI["urn:cts:latinLit:phi1294"].metadata["groupname"]["eng"], "Martial")
        self.assertEqual(TI["urn:cts:latinLit:phi1294"].metadata["groupname"]["lat"], "Martialis")
        self.assertEqual(TI["urn:cts:latinLit:phi1294.phi002"].metadata["title"]["eng"], "Epigrammata")
        self.assertEqual(TI["urn:cts:latinLit:phi1294.phi002"].metadata["title"]["fre"], "Epigrammes")
        self.assertEqual(TI["urn:cts:latinLit:phi1294.phi002.perseus-lat2"].metadata["label"]["eng"], "Epigrammata Label")
        self.assertEqual(TI["urn:cts:latinLit:phi1294.phi002.perseus-lat2"].metadata["label"]["fre"], "Epigrammes Label")
        self.assertEqual(TI["urn:cts:latinLit:phi1294.phi002.perseus-lat2"].metadata["description"]["fre"], "G. Heraeus")
        self.assertEqual(TI["urn:cts:latinLit:phi1294.phi002.perseus-lat2"].metadata["description"]["eng"], "W. Heraeus")

    def test_export(self):
        ed = """<ti:edition urn='urn:cts:latinLit:phi1294.phi002.perseus-lat2' workUrn='urn:cts:latinLit:phi1294.phi002' xmlns:ti='http://chs.harvard.edu/xmlns/cts'>
<ti:label xml:lang='eng'>Epigrammata Label</ti:label>
<ti:label xml:lang='fre'>Epigrammes Label</ti:label>
<ti:description xml:lang='eng'>W. Heraeus</ti:description>
<ti:description xml:lang='fre'>G. Heraeus</ti:description>
<ti:online docname='/db/apps/canonical-latinLit/data/phi1294/phi002/phi1294.phi002.perseus-lat2.xml'>
<ti:validate schema='tei-epidoc.rng'/>
<ti:namespaceMapping abbreviation='tei' nsURI='http://www.tei-c.org/ns/1.0'/>
<ti:citationMapping>
<ti:citation label='book' xpath='/tei:div[@n="?"]' scope='/tei:TEI/tei:text/tei:body/tei:div'>
<ti:citation label='poem' xpath='/tei:div[@n="?"]' scope='/tei:TEI/tei:text/tei:body/tei:div/tei:div[@n="?"]'>
<ti:citation label='line' xpath='/tei:l[@n="?"]' scope='/tei:TEI/tei:text/tei:body/tei:div/tei:div[@n="?"]/tei:div[@n="?"]'></ti:citation>
</ti:citation>
</ti:citation>
</ti:citationMapping>
</ti:online
></ti:edition>""".replace("\n", "")

        tr = """<ti:translation xml:lang='eng' urn='urn:cts:latinLit:phi1294.phi002.perseus-eng2' workUrn='urn:cts:latinLit:phi1294.phi002' xmlns:ti='http://chs.harvard.edu/xmlns/cts'>
<ti:label xml:lang='eng'>Epigrammata</ti:label>
<ti:description xml:lang='eng'>M. Valerii Martialis Epigrammaton libri / recognovit W. Heraeus</ti:description>
<ti:online></ti:online>
</ti:translation>""".replace("\n", "")

        wk = """<ti:work urn='urn:cts:latinLit:phi1294.phi002' groupUrn='urn:cts:latinLit:phi1294' xmlns:ti='http://chs.harvard.edu/xmlns/cts'>
<ti:title xml:lang='eng'>Epigrammata</ti:title>
<ti:title xml:lang='fre'>Epigrammes</ti:title>""" + tr + ed + """</ti:work>""".replace("\n", "")

        tg = """<ti:textgroup urn='urn:cts:latinLit:phi1294' xmlns:ti='http://chs.harvard.edu/xmlns/cts'>
<ti:groupname xml:lang='eng'>Martial</ti:groupname>
<ti:groupname xml:lang='lat'>Martialis</ti:groupname>""" + wk + """</ti:textgroup>""".replace("\n", "")

        t = """<ti:TextInventory tiid='annotsrc' xmlns:ti='http://chs.harvard.edu/xmlns/cts'>""" + tg + """</ti:TextInventory>""".replace("\n", "").strip("\n")

        ti = TextInventory(resource=t, id="annotsrc")
        self.assertXmlEquivalentOutputs(*compareSTR(str(ti), t))

        # Test individual :
        self.assertXmlEquivalentOutputs(*compareSTR(str(ti["urn='urn:cts:latinLit:phi1294"]), tg))
        self.assertXmlEquivalentOutputs(*compareSTR(str(ti["urn='urn:cts:latinLit:phi1294.phi002"]), wk))
        self.assertXmlEquivalentOutputs(*compareSTR(str(ti["urn='urn:cts:latinLit:phi1294.phi002.perseus-eng2"]), tr))
        self.assertXmlEquivalentOutputs(*compareSTR(str(ti["urn='urn:cts:latinLit:phi1294.phi002.perseus-lat2"]), ed))

        # Test export :
        self.assertXmlEquivalentOutputs(*compareXML(ti.export(), t))
        self.assertXmlEquivalentOutputs(*compareXML(ti["urn='urn:cts:latinLit:phi1294"].export(), tg))
        self.assertXmlEquivalentOutputs(*compareXML(ti["urn='urn:cts:latinLit:phi1294.phi002"].export(), wk))
        self.assertXmlEquivalentOutputs(*compareXML(ti["urn='urn:cts:latinLit:phi1294.phi002.perseus-eng2"].export(), tr))
        self.assertXmlEquivalentOutputs(*compareXML(ti["urn='urn:cts:latinLit:phi1294.phi002.perseus-lat2"].export(), ed))

    def test_export_to_text(self):
        """ Test export to Text object """
        TI = TextInventory(resource=self.getCapabilities, id="annotsrc")
        ti_text = TI["urn:cts:latinLit:phi1294.phi002.perseus-lat2"]

        txt_text = ti_text.export(output=MyCapytain.resources.proto.text.Text)
        self.assertEqual(str(txt_text.urn), "urn:cts:latinLit:phi1294.phi002.perseus-lat2")
        self.assertEqual(txt_text.metadata["groupname"]["eng"], "Martial")  # Check inheritance of textgroup metadata
        self.assertEqual(txt_text.metadata["title"]["eng"], "Epigrammata")  # Check inheritance of work metadata
        self.assertEqual(txt_text.metadata["title"]["fre"], "Epigrammes")  # Check inheritance of work metadata
        self.assertEqual(txt_text.metadata["description"]["fre"], "G. Heraeus")  # Check inheritance of work metadata
        self.assertEqual(txt_text.citation, ti_text.citation)
        self.assertEqual(txt_text.citation.scope, "/tei:TEI/tei:text/tei:body/tei:div")

    def test_partial_str(self):
        ti = TextInventory(resource=self.t, id="annotsrc")

        e = deepcopy(ti["urn='urn:cts:latinLit:phi1294.phi002.perseus-lat2"])
        e.urn = None
        self.assertXmlEquivalentOutputs(
            *compareSTR(
                    str(e), 
                    self.ed.replace("urn='urn:cts:latinLit:phi1294.phi002.perseus-lat2' ", "")
                )
            )
        e.parents = ()
        self.assertXmlEquivalentOutputs(
            *compareSTR(
                    str(e), 
                    self.ed.replace(
                        "urn='urn:cts:latinLit:phi1294.phi002.perseus-lat2' ", ""
                    ).replace(
                        "workUrn='urn:cts:latinLit:phi1294.phi002' ", ""
                    )
                )
            )

        wk = deepcopy(ti["urn='urn:cts:latinLit:phi1294.phi002"])
        wk.urn = None
        self.assertXmlEquivalentOutputs(
            *compareSTR(
                    str(wk), 
                    self.wk.replace("urn='urn:cts:latinLit:phi1294.phi002' ", "")
                )
            )
        wk.parents = ()
        self.assertXmlEquivalentOutputs(
            *compareSTR(
                    str(wk), 
                    self.wk.replace(
                        "urn='urn:cts:latinLit:phi1294.phi002' ", ""
                    ).replace(
                        "groupUrn='urn:cts:latinLit:phi1294' ", ""
                    )
                )
            )

        tg = deepcopy(ti["urn='urn:cts:latinLit:phi1294"])
        tg.urn = None
        self.assertXmlEquivalentOutputs(
            *compareSTR(
                    str(tg), 
                    self.tg.replace("urn='urn:cts:latinLit:phi1294' ", "")
                )
            )

        ti = deepcopy(ti)
        ti.id = None
        self.assertXmlEquivalentOutputs(
            *compareSTR(
                    str(ti), 
                    self.t.replace("tiid='annotsrc' ", "")
                )
            )

class TestCitation(unittest.TestCase):
    def test_empty(self):
        a = Citation(name="none")
        self.assertEqual(str(a), "")